Option Explicit

' Miscellaneous ACO-related utility functions.


'------------------
' Public functions
'------------------

Public Function AcoFirstInPack(acoPack)
	Call TraceEnter("AcoFirstInPack", SzTraceAco(acoPack))
	
	If acoPack.citemContents > 0 Then
		Set AcoFirstInPack = acoPack.coacoContents(0)
	Else
		Set AcoFirstInPack = Nothing
	End If
	
	Call TraceExit("AcoFirstInPack", SzTraceAco(AcoFirstInPack))
End Function

Public Function AcoNextInPack(aco)
	Call TraceEnter("AcoNextInPack", SzTraceAco(aco))
	
	Dim iitem : iitem = aco.iitem + 1
	If iitem < aco.acoContainer.citemContents Then
		Set AcoNextInPack = aco.acoContainer.coacoContents(iitem)
	Else
		Set AcoNextInPack = Nothing
	End If
	
	Call TraceExit("AcoNextInPack", SzTraceAco(AcoNextInPack))
End Function

' Return True if the given object is a door.
Public Function FAcoIsDoor(aco)
	Call TraceEnter("FAcoIsDoor", SzTraceAco(aco))
	
	FAcoIsDoor = ((aco.oty And otyDoor) <> 0)
	
	Call TraceExit("FAcoIsDoor", CStr(FAcoIsDoor))
End Function

' Return True if the given object is a portal.
Private Function FAcoIsPortal(aco)
	Call TraceEnter("FAcoIsPortal", SzTraceAco(aco))
	
	FAcoIsPortal = ((aco.oty And otyPortal) <> 0)
	
	Call TraceExit("FAcoIsPortal", CStr(FAcoIsPortal))
End Function

' Return True if the given object is stackable.
Public Function FStackableAco(aco)
	Call TraceEnterEx("FStackableAco", SzTraceAco(aco), tefAcoUtil)
	
	FStackableAco = (aco.citemMaxStack > 1)
	
	Call TraceExit("FStackableAco", CStr(FStackableAco))
End Function


' Return a string representation of the given ACO suitable for use with 
' TraceEnter/TraceExit.
Public Function SzTraceAco(aco)
	
	If aco Is Nothing Then
		SzTraceAco = "Nothing"
	Else
		SzTraceAco = SzTraceOid(aco.oid) + " " + SzQuote(aco.szName)
	End If
	
End Function

' Return a string representation of the given COACO suitable for use with 
' TraceEnter/TraceExit.
Public Function SzTraceCoaco(coaco)
	
	If coaco Is Nothing Then
		SzTraceCoaco = "Nothing"
	Else
		SzTraceCoaco = SzNItems(coaco.Count, "item")
	End If
	
End Function

' Return a string representation of the given maploc suitable for use with 
' TraceEnter/TraceExit.
Public Function SzTraceMaploc(maploc)
	
	If maploc Is Nothing Then
		SzTraceMaploc = "Nothing"
	Else
		SzTraceMaploc = maploc.sz(3)
	End If
	
End Function

' Return a string representation of the given EQM suitable for use with 
' TraceEnter/TraceExit.
Public Function SzTraceEqm(eqm)
	SzTraceEqm = SzHex8(eqm)
End Function

' Return a string representation of the given MCM suitable for use with 
' TraceEnter/TraceExit.
Public Function SzTraceMcm(mcm)
	SzTraceMcm = SzHex8(mcm)
End Function

' Return a string representation of the given OID suitable for use with 
' TraceEnter/TraceExit.
Public Function SzTraceOid(oid)
	SzTraceOid = SzHex8(oid)
End Function

' Return a string representation of the given OTY suitable for use with 
' TraceEnter/TraceExit.
Public Function SzTraceOty(oty)
	SzTraceOty = SzHex8(oty)
End Function


