Option Explicit
Class AssessCls

' Assessment utilities.  All functions should be accessed as methods of the 
' global Assess object (e.g. Assess.FLockedAco(aco)).  The Assess object is 
' created and initialized automatically on script load.


'-------------------
' Class boilerplate
'-------------------

Public tef

Private Sub Class_Initialize()
	Call TraceEnter("Assess.Initialize", "")
	
	cOaiAssessAco = 0
	
	Call TraceExit("Assess.Initialize", "")
End Sub

Private Sub Class_Terminate()
	Call TraceEnter("Assess.Terminate", "")
	
	Call TraceExit("Assess.Terminate", "")
End Sub


'------------
' Properties
'------------

Public Property Get fUserAssess()
	fUserAssess = (cOaiAssessAco = 0)
End Property


'---------
' Methods
'---------

' Given an ACO, return assessment info not older than cmsecMax milliseconds.  
' If the ACO has current assessment info younger than that, just return it.  
' Otherwise assess the object, wait for the results, and return the fresh info.  
' If cmsecMax is -1, then we don't care how old the info is, and a new assess 
' is required only if the ACO has never been assessed since its creation.
Public Function OaiFromAco(aco, cmsecMax)
	Call TraceEnterEx("Assess.OaiFromAco", _
	  SzTraceAco(aco) + ", " + CStr(cmsecMax), tef)
	Do
	
	If aco is Nothing Then
		Set OaiFromAco = Nothing
		Exit Do
	End If
		
	If Not (aco.oai Is Nothing) Then
		If cmsecMax < 0 Or aco.oai.cmsecSinceAssess <= cmsecMax Then
			Set OaiFromAco = aco.oai
			Exit Do
		End If
	End If
	
	Dim itry
	For itry = 0 To 2
		If Not aco.fExists Then
			Set OaiFromAco = Nothing
			Exit For
		End If
		
		Set OaiFromAco = OaiAssessAco(aco, 5)
		If Not (OaiFromAco Is Nothing) Then Exit For
	Next
	
	Loop While False
	Call TraceExit("Assess.OaiFromAco", SzTraceOai(OaiFromAco))
End Function

' Assess the currently selected object.  Wait up to csecTimeout seconds for the 
' results of the assessment (the OnAssessItem event), and return the received 
' information.
Public Function OaiAssessSel(csecTimeout)
	Call TraceEnterEx("Assess.OaiAssessSel", CStr(csecTimeout), tef)
	
	Set OaiAssessSel = OaiAssessAco(skapi.acoSelected, csecTimeout)
	
	Call TraceExit("Assess.OaiAssessSel", SzTraceOai(OaiAssessSel))
End Function

' Assess the object specified by aco.  Wait up to csecTimeout seconds for the 
' results of the assessment (the OnAssessItem event), and return the received 
' information.
Public Function OaiAssessAco(aco, csecTimeout)
	Call TraceEnterEx("Assess.OaiAssessAco", _
	  SzTraceAco(aco) + ", " + CStr(csecTimeout), tef)
	Call PushUnwind(Me)
	cOaiAssessAco = cOaiAssessAco + 1
	
'	Process queued events to get any pending assessments out of the way.
	Call skapi.WaitEvent()
	
	If aco Is Nothing Then
		Set OaiAssessAco = Nothing
	Else
	'	Assess the item.
		Call skapi.AssessAco(aco)
		
	'	Wait for the results.
		Dim oaiPrev : Set oaiPrev = aco.oai
		Dim cmsecTimeout : cmsecTimeout = csecTimeout * 1000
		Dim timer : Set timer = skapi.TimerNew()
		Do
			Call Pauser.Check()
			If timer.cmsec > cmsecTimeout Then Exit Do
			Call skapi.WaitEvent(100)
		Loop While aco.oai Is oaiPrev
		
	'	Return the info.
		If aco.oai Is oaiPrev Then
		'	Sometimes assessment gets stuck on a particular item.
		'	Assess something else to try to shake it loose.
			Call skapi.AssessAco(skapi.acoChar)
			
			Set OaiAssessAco = Nothing
		Else
			Set OaiAssessAco = aco.oai
		End If
	End If
	
	cOaiAssessAco = cOaiAssessAco - 1
	Call PopUnwind(Me)
	Call TraceExit("Assess.OaiAssessAco", SzTraceOai(OaiAssessAco))
End Function

' Assess the given aco and return its inscription if any.
Public Function SzInscriptionFromAco(aco)
	Call TraceEnterEx("Assess.SzInscriptionFromAco", SzTraceAco(aco), tef)
	
	SzInscriptionFromAco = ""
	
	If (aco.oty And otyInscribable) <> 0 Then
		Dim oai : Set oai = OaiFromAco(aco, -1)
		If Not (oai Is Nothing) Then
			If Not (oai.ibi Is Nothing) Then
				SzInscriptionFromAco = oai.ibi.szInscription
			End If
		End If
	End If
	
	Call TraceExit("Assess.SzInscriptionFromAco", SzQuote(SzInscriptionFromAco))
End Function

' Assess the given aco, presumed to be a portal, and return its approximate 
' destination coords if available.
Public Function MaplocDstOfPortal(aco)
	Call TraceEnter("Assess.MaplocDstOfPortal", SzTraceAco(aco))
	
	Dim oai : Set oai = OaiFromAco(aco, -1)
	Call Assert(Not (oai Is Nothing), "Assess failed.")
	Call Assert(Not (oai.ipi Is Nothing) <> 0, "No portal info.")
	
	Dim szDst : szDst = oai.ipi.szDest
	Call TraceLine(szDst)
	Dim rex : Set rex = New RegExp
	rex.Pattern = "\D*(\d+(\.\d)?(N|S)),?\s*(\d+(\.\d)?(E|W)).*"
	rex.Global = False
	rex.IgnoreCase = True
	Dim comatch : Set comatch = rex.Execute(szDst)
	If comatch.Count > 0 Then
		szDst = comatch(0).SubMatches(0) + " " + comatch(0).SubMatches(3)
		Set MaplocDstOfPortal = skapi.MaplocFromSz(szDst)
	Else
		Set MaplocDstOfPortal = Nothing
	End If
	
	Call TraceExit("Assess.MaplocDstOfPortal", SzTraceMaploc(MaplocDstOfPortal))
End Function

' Return True if the given aco (door, chest, or whatnot) is locked.
Public Function FLockedAco(aco)
	Call TraceEnterEx("Assess.FLockedAco", SzTraceAco(aco), tef)
	
'	REVIEW: Apparently there are doors without this flag:
'	If (aco.mcm And mcmLockable) = 0 Then
'		FLockedAco = False
'	Else
	If (aco.ocm And (ocmPlayerCorpse Or ocmMonsterCorpse)) <> 0 Then
		FLockedAco = False
	Else
		FLockedAco = False
		Dim oai : Set oai = OaiAssessAco(aco, 5)
		If oai Is Nothing Then Call RaiseTimeout("Assess.FLockedAco")
		FLockedAco = oai.ibi.fLocked
	End If
	
	Call TraceExit("Assess.FLockedAco", CStr(FLockedAco))
End Function

' Return True if the given aco (door, chest, or whatnot) is open.
Public Function FOpenAco(aco)
	Call TraceEnterEx("Assess.FOpenAco", SzTraceAco(aco), tef)
	
	If aco Is skapi.acoChar Then
		FOpenAco = True
	ElseIf (aco.oty And otyContainer) <> 0 And aco.coacoContents Is Nothing Then
		FOpenAco = False
	Else
		FOpenAco = False
		Dim oai : Set oai = OaiAssessAco(aco, 5)
		If oai Is Nothing Then Call RaiseTimeout("Assess.FOpenAco")
		FOpenAco = oai.ibi.fOpen
	End If
	
	Call TraceExit("Assess.FOpenAco", CStr(FOpenAco))
End Function

' If the given aco (door, chest, or whatnot) is locked, unlock it.  acoKey is 
' the necessary key; if acoKey is Nothing, and aco is locked, then raise an 
' error.
Public Function EnsureUnlockedAco(aco, acoKey, csecTimeout)
	Call TraceEnterEx("Assess.EnsureUnlockedAco", _
	  SzTraceAco(aco) + ", " + SzTraceAco(acoKey) + ", " + CStr(csecTimeout), tef)
	
	If FLockedAco(aco) Then
		If acoKey Is Nothing Then
			Call RaiseCmr(aco.szName + " is locked.")
		End If
		Call acoKey.UseOnAco(aco)
		Dim timer : Set timer = skapi.TimerNew()
		Do While FLockedAco(aco)
			Call Pauser.Check()
			Call CheckTimeout(timer, csecTimeout, "Assess.EnsureUnlockedAco")
			Call skapi.WaitEvent(300, wemFullTimeout)
		Loop
	End If
	
	Call TraceExit("Assess.EnsureUnlockedAco", "")
End Function

' If fOpen is True, and the given aco (door, chest, or whatnot) is closed, 
' then open it.  (It's assumed to be unlocked already.)  If fOpen is False, 
' and aco is open, then close it.
Public Function EnsureFOpenAco(aco, fOpen, csecTimeout)
	Call TraceEnterEx("Assess.EnsureFOpenAco", _
	  SzTraceAco(aco) + ", " + CStr(fOpen) + ", " + CStr(csecTimeout), tef)
	
	If FOpenAco(aco) <> fOpen Then
		Call aco.Use()
		Dim timer : Set timer = skapi.TimerNew()
		Do Until FOpenAco(aco) = fOpen
			Call Pauser.Check()
			Call CheckTimeout(timer, csecTimeout, "Assess.EnsureFOpenAco")
			Call skapi.WaitEvent(300, wemFullTimeout)
		Loop
	End If
	
	Call TraceExit("Assess.EnsureFOpenAco", "")
End Function

' Assess the given aco and return the time in seconds until its mana charge 
' runs out.
Public Function CsecTillRechargeAco(aco)
	Call TraceEnterEx("Assess.CsecTillRechargeAco", SzTraceAco(aco), tef)
	
	CsecTillRechargeAco = 0
	
	Dim oai : Set oai = OaiAssessAco(aco, 5)
	If Not (oai Is Nothing) Then
		If Not (oai.iei Is Nothing) Then
			CsecTillRechargeAco = oai.iei.manaCur * -oai.iei.csecPerMana
		End If
	End If
	
	Call TraceExit("Assess.CsecTillRechargeAco", CStr(Round(CsecTillRechargeAco)))
End Function

' Return a string representation of the given oai suitable for use with 
' TraceEnter/TraceExit.
Public Function SzTraceOai(oai)
	
	If oai Is Nothing Then
		SzTraceOai = "Nothing"
	Else
		SzTraceOai = ""
	End If
	
End Function


'----------------
' Unwind handler
'----------------

' Remove event handlers and such in case of error.
Public Function Unwind(erh)
	Call TraceEnterEx("Assess.Unwind", "", tefEnable)
	
	cOaiAssessAco = cOaiAssessAco - 1
	
	Call TraceExit("Assess.Unwind", "")
End Function


'----------
' Privates
'----------

Private cOaiAssessAco


End Class


' Instantiate the global object.
Public Assess
Set Assess = New AssessCls


