Option Explicit
Class WayptJumpCls
' Jumpoff waypoint class (as in Subway).


'---------------------------------------
' Type-specific properties and methods:
'---------------------------------------

Public maplocIn			' The location of the jumpoff point.
Public maplocOut		' The location of the landing point.
Public healthReq		' Minimum health needed to survive this jump.

' Initialize this waypoint from the given parameters.
Public Function Init(maplocInA, maplocOutA, healthReqA)
	Call TraceEnterEx("WayptJump.Init", SzTraceMaploc(maplocInA) + ", " + _
	  SzTraceMaploc(maplocOutA) + ", " + CStr(healthReqA), Xnav.tef)
	
	Set maplocIn = maplocInA
	Set maplocOut = maplocOutA
	healthReq = healthReqA
	
	Call TraceExit("WayptJump.Init", "")
End Function


'----------------------------------
' Required properties and methods:
'----------------------------------

' The XML tag name of this waypoint type.
Public Property Get szTag()
	szTag = "Jump"
End Property

' True if this leg should be walked, false if it should be run.
' Xnav handles initialization of this property, but each waypoint class must define it.
Public fWalk

' maplocIn and maplocOut are listed under "Type-specific properties and 
' methods" above.

' The arrival distance of this waypoint in MU, i.e. how accurately we need to approach it 
' before performing the arrival action (if any) and passing on to the next waypoint.
Public Property Get distArrive()
	distArrive = 1 * Xnav.distMeter
End Property

' The estimated time in seconds needed for this waypoint's arrival action.
Public Property Get csecArrive()
	csecArrive = 10	' Figure 10 seconds for the fall.
End Property

' Perform this waypoint's arrival action.
Public Function Arrive()
	Call TraceEnterEx("WayptJump.Arrive", SzTraceMaploc(maplocIn), Xnav.tef)
	
	If skapi.healthCur < healthReq Then
	'	Replace the following line with your custom healing code.
		Call RaiseCmr("You're not healthy enough for this jump.")
	End If
	
'	Face maplocOut.
	Call Xnav.TurnTowardMaploc(maplocOut)
	
'	Jump.
	Call skapi.Keys(cmidJump, kmoDown)
	If maplocOut.z - maplocIn.z > 1 Then
		Call skapi.Sleep(1500)
	ElseIf maplocIn.z - maplocOut.z < 3 Then
		Call skapi.Sleep(300)
	End If
	Call skapi.Keys(cmidWalkForward, kmoDown)
	Call skapi.Keys(cmidJump, kmoUp)
	Call skapi.Keys(cmidWalkForward, kmoUp)
	
'	Wait for landing.
	Dim timer : Set timer = skapi.TimerNew()
	Do While Abs(skapi.maplocCur.z - maplocOut.z) > 1
		If timer.cmsec > 10000 Then Call RaiseCmr("Jump failed.")
		Call skapi.WaitEvent(100, wemFullTimeout)
	Loop
	
	Call TraceExit("WayptJump.Arrive", "")
End Function

' Generate a new waypoint representing the reversal of this waypoint.
' Used by Xroute.XrouteReverse.
Public Function WayptReverse()
	Call TraceEnterEx("WayptJump.WayptReverse", "", Xnav.tef)
	
	If Abs(maplocIn.z - maplocOut.z) < 2 Then
	'	Horizontal jumps are reversible.
		Set WayptReverse = WayptJumpNew()
		Call WayptReverse.Init(maplocOut, maplocIn, healthReq)
	Else
	'	Long vertical jumps are not reversible.
		Set WayptReverse = Nothing
	End If
	
	Call TraceExit("WayptJump.WayptReverse", "")
End Function

' Initialize this waypoint from skapi.maplocCur and, if relevant, skapi.acoSelected.
' Used by XrouteEdit.
Public Function FromMaplocCur()
	Call TraceEnterEx("WayptJump.FromMaplocCur", "", Xnav.tef)
	
	Call Init(skapi.maplocCur, skapi.maplocCur, 50)
	
	Call TraceExit("WayptJump.FromMaplocCur", "")
End Function

' Initialize this waypoint from an XML element.
Public Function FromXml(elem)
	Call TraceEnterEx("WayptJump.FromXml", SzQuote(elem.xml), Xnav.tef)
	
	healthReq = LGetAttr(elem, "healthReq", 0)
	Set maplocIn = Xnav.MaplocOfElem(elem.SelectSingleNode("maplocIn"))
	Set maplocOut = Xnav.MaplocOfElem(elem.SelectSingleNode("maplocOut"))
	
	Call TraceExit("WayptJump.FromXml", "")
End Function

' Save type-specific waypoint parameters to the given XML element.
' If fVerbose is True, all waypoint parameters should be made explicit in the XML.  
' (This is needed by XrouteEdit.)  If fVerbose is False, parameters with default 
' values may be omitted.
Public Function ToXml(elem, fVerbose)
	Call TraceEnterEx("WayptJump.ToXml", CStr(fVerbose), Xnav.tef)
	
	Call SetLAttr(elem, "healthReq", healthReq)
	
	Dim elemT : Set elemT = elem.ownerDocument.CreateElement("maplocIn")
	Call Xnav.SetElemMaploc(elemT, maplocIn, True, False)
	Call elem.AppendChild(elemT)
	
	Set elemT = elem.ownerDocument.CreateElement("maplocOut")
	Call Xnav.SetElemMaploc(elemT, maplocOut, True, False)
	Call elem.AppendChild(elemT)
	
	Call TraceExit("WayptJump.ToXml", "")
End Function


'-----------
' Privates:
'-----------

Private Sub Class_Initialize()
	
	Set maplocIn = Nothing
	Set maplocOut = Nothing
	healthReq = 0
	
End Sub


End Class


' Instantiator function.
Public Function WayptJumpNew()
	Set WayptJumpNew = New WayptJumpCls
End Function

' Register the waypoint type.
Call Xnav.RegisterType("Jump", "WayptJumpNew")


