Option Explicit
Class XnavCls


'-------------------
' Class boilerplate
'-------------------

Public tef

Private Sub Class_Initialize()
	Call TraceEnter("Xnav.Initialize", "")
	
	Set dictTagFunc = CreateObject("Scripting.Dictionary")
	dictTagFunc.CompareMode = vbTextCompare
	
	Set xdocRecent = Nothing
	Set dictSzXrouteRecent = Nothing
	
	crouteInProgress = 0
	
	Call TraceExit("Xnav.Initialize", "")
End Sub

Private Sub Class_Terminate()
	Call TraceEnter("Xnav.Terminate", "")
	
	Call TraceExit("Xnav.Terminate", "")
End Sub


'-------------------
' Public properties
'-------------------

Public Property Get distMeter()
	distMeter = 1/240			' One map unit = 240 meters.
End Property
Public Property Get distPace()
	distPace = distMeter/2	' REVIEW: Measure this.
End Property

Public Property Get vWalk()
	vWalk = 0.0108		' Walking speed in MU/s.
End Property
Public Property Get vRunTypical()
	vRunTypical = 0.035			' Typical running speed in MU/s.
End Property

Public Property Get dheadPerSec()
	dheadPerSec = 90			' Turning speed in degrees/s while stopped or walking.
End Property
Public Property Get dheadPerSecRun()
	dheadPerSecRun = 133			' Turning speed in degrees/s while running.
End Property
Public Property Get dheadRadPerSec()
	dheadRadPerSec = 1.57			' Turning speed in radians/s while stopped or walking.
End Property
Public Property Get dheadRadPerSecRun()
	dheadRadPerSecRun = 2.32			' Turning speed in radians/s while running.
End Property
Public Property Get pi()
	pi = 3.14159
End Property


'----------------
' Public methods
'----------------

' Move your character to the given latitude and longitude by the most direct 
' route.
Public Function GoToLatLng(lat, lng)
	Call TraceEnterEx("Xnav.GoToLatLng", _
	  FormatNumber(lat, 3) + ", " + FormatNumber(lng, 3), tef)
	
	Call GoToMaploc(skapi.MaplocFromLatLng(lat, lng))
	
	Call TraceExit("Xnav.GoToLatLng", "")
End Function

' Move your character to the given location by the most direct route.  This is 
' equivalent to Xnav.XrouteFromMaploc(maploc).Go().
Public Function GoToMaploc(maploc)
	Call TraceEnterEx("Xnav.GoToMaploc", SzTraceMaploc(maploc), tef)
	
	Call XrouteFromMaploc(maploc).Go()
	
	Call TraceExit("Xnav.GoToMaploc", "")
End Function

' Move your character to the given location by the most direct route.  This is 
' equivalent to Xnav.XrouteFromMaploc(maploc).Go().
Public Function GoToMaplocEx(maploc, distArrive)
	Call TraceEnterEx("Xnav.GoToMaplocEx", SzTraceMaploc(maploc) + ", " + FormatNumber(distArrive), tef)
	
	Dim xroute : Set xroute = XrouteNew()
	Dim waypt
	If distArrive > 3 * distMeter Then
	'	Run to the target.
		Set waypt = WayptFromTag("Simple")
		Call waypt.Init(maploc, distArrive)
		waypt.fWalk = False
		Call xroute.InsertWaypt(0, waypt)
	Else
	'	Walk to the target.
		Set waypt = WayptFromTag("Simple")
		Call waypt.Init(maploc, distArrive)
		waypt.fWalk = True
		Call xroute.InsertWaypt(0, waypt)
		If skapi.maplocCur.Dist2DToMaploc(maploc) > 6 * distMeter Then
		'	Run most of the way first.
			Set waypt = WayptFromTag("Simple")
			Call waypt.Init(maploc, 4 * distMeter)
			waypt.fWalk = False
			Call xroute.InsertWaypt(0, waypt)
		End If
	End If
	Call xroute.Go()
	
	Call TraceExit("Xnav.GoToMaplocEx", "")
End Function

Public Function MaplocZeroInOnMaploc(maplocTarget, distArrive)
	Call TraceEnterEx("Xnav.MaplocZeroInOnMaploc", _
	  SzTraceMaploc(maplocTarget) + ", " + FormatNumber(distArrive), tef)
	
	Dim maplocCur
	Do
		Set maplocCur = skapi.maplocCur
		If maplocCur.Dist2DToMaploc(maplocTarget) <= distArrive Then Exit Do
		Call TurnTowardMaploc(maplocTarget)
		Call GoToMaplocEx(maplocTarget, distArrive)
		Call Pauser.Check()
	Loop
	
	Set MaplocZeroInOnMaploc = maplocCur
	
	Call TraceExit("Xnav.MaplocZeroInOnMaploc", SzTraceMaploc(MaplocZeroInOnMaploc))
End Function

' Stop the execution of any route in progress and return control to the 
' caller of xroute.Go.  Has no effect if there is no route in progress.
Public Function StopNav()
	Call TraceEnterEx("Xnav.StopNav", "", tef)
	
	fStopNav = True
	
	Call TraceExit("Xnav.StopNav", "")
End Function


' Turn your character to face the given maploc, but don't go there.
Public Function TurnTowardMaploc(maploc)
	Call TraceEnterEx("Xnav.TurnTowardMaploc", SzTraceMaploc(maploc), tef)
	
	Call TurnToHead(skapi.maplocCur.HeadToMaploc(maploc))
	
	Call TraceExit("Xnav.TurnTowardMaploc", "")
End Function

' Turn your character to face the given compass heading (in degrees, measured 
' clockwise from north).
Public Function TurnToHead(head)
	Call TraceEnterEx("Xnav.TurnToHead", FormatNumber(head, 1), tef)

	Call skapi.TurnToHead(head)
	Call skapi.WaitEvent(5000, wemSpecific, evidOnNavStop)
	
	Call TraceExit("Xnav.TurnToHead", "")
End Function


' Create and return a new, empty Xroute object.
Public Function XrouteNew()
	Call TraceEnterEx("Xnav.XrouteNew", "", tef)
	
	Set XrouteNew = New XrouteCls
	
	Call TraceExit("Xnav.XrouteNew", SzTraceObj(XrouteNew))
End Function

' Create and return a direct, straight-line route (i.e. a route of just one 
' point) to the given location.
Public Function XrouteFromMaploc(maploc)
	Call TraceEnterEx("Xnav.XrouteFromMaploc", SzTraceMaploc(maploc), tef)
	
	Dim xroute : Set xroute = XrouteNew()
	Call xroute.InsertMaploc(iNil, maploc)
	Set XrouteFromMaploc = xroute
	
	Call TraceExit("Xnav.XrouteFromMaploc", "")
End Function

' Create and return a new route object from the contents of the named file.
Public Function XrouteFromFile(szFile)
	Call TraceEnterEx("Xnav.XrouteFromFile", SzQuote(szFile), tef)
	
	Dim xroute : Set xroute = XrouteNew()
	Call xroute.FromFile(szFile, False)
	Set XrouteFromFile = xroute
	
	Call TraceExit("Xnav.XrouteFromFile", SzTraceObj(XrouteFromFile))
End Function

' Create and return a new route object from the given XML text.
Public Function XrouteFromSz(szXml, fInterpolate)
	Call TraceEnterEx("Xnav.XrouteFromSz", CStr(fInterpolate), tef)
	
	Dim xroute : Set xroute = XrouteNew()
	Call xroute.FromSz(szXml, fInterpolate)
	Set XrouteFromSz = xroute
	
	Call TraceExit("Xnav.XrouteFromSz", SzTraceObj(XrouteFromSz))
End Function

' Create and return a new route object from the given XML element.
Public Function XrouteFromXml(elem, fInterpolate)
	Call TraceEnterEx("Xnav.XrouteFromXml", CStr(fInterpolate), tef)
	
	Dim xroute : Set xroute = XrouteNew()
	Call xroute.FromXml(elem, fInterpolate)
	Set XrouteFromXml = xroute
	
	Call TraceExit("Xnav.XrouteFromXml", SzTraceObj(XrouteFromXml))
End Function

Public Function DictSzXrouteFromFile(szFile)
	Call TraceEnterEx("Xnav.DictSzXrouteFromFile", SzQuote(szFile), tef)
	
	Dim xdoc : Set xdoc = XdocLoadFile(szFile)
	
	Set DictSzXrouteFromFile = DictSzXrouteFromXdoc(xdoc)
	
	Call TraceExit("Xnav.DictSzXrouteFromFile", _
	  SzNItems(DictSzXrouteFromFile.Count, "route"))
End Function

Public Function DictSzXrouteFromXdoc(xdoc)
	Call TraceEnterEx("Xnav.DictSzXrouteFromXdoc", SzTraceXdoc(xdoc), tef)
	
	Set xdocRecent = xdoc
	Set dictSzXrouteRecent = CreateObject("Scripting.Dictionary")
	dictSzXrouteRecent.CompareMode = vbTextCompare
	
	Dim sel, elemXroute, iroute
	Set sel = xdoc.SelectNodes("//Xroute")
	iroute = 0
	For Each elemXroute In sel
		Dim xroute : Set xroute = XrouteNew()
		Call xroute.FromXml(elemXroute, False)
		If xroute.szName = "" Then
		'	Assign default names to anonymous routes.
			If iroute = 0 Then
				xroute.szName = "Default route"
			Else
				xroute.szName = "Route " + CStr(iroute + 1)
			End If
		End If
		Set dictSzXrouteRecent(xroute.szName) = xroute
		iroute = iroute + 1
	Next
	
	Set DictSzXrouteFromXdoc = dictSzXrouteRecent
	
	Call TraceExit("Xnav.DictSzXrouteFromXdoc", _
	  SzNItems(DictSzXrouteFromXdoc.Count, "route"))
End Function

' Create and return an XML document representing an empty Xroute file.
Public Function XdocNewXrouteFile()
	Call TraceEnterEx("Xnav.XdocNewXrouteFile", "", tef)
	
	Set XdocNewXrouteFile = XdocNew("XrouteFile")
	
	Call TraceExit("Xnav.XdocNewXrouteFile", SzTraceXdoc(XdocNewXrouteFile))
End Function


' Register a waypoint type.  szTag is the type's unique XML tag name.  szFunc 
' is the name of the global function used to create an instance of that type.
Public Function RegisterType(szTag, szFunc)
	Call TraceEnterEx("Xnav.RegisterType", SzQuote(szTag) + ", " + SzQuote(szFunc), tef)
	
	Call Assert(Not dictTagFunc.Exists(szTag), "Duplicate waypoint tag.")
	
	dictTagFunc(szTag) = szFunc
	
	Call TraceExit("Xnav.RegisterType", "")
End Function

' Return the ith registered waypoint tag, or the empty string is there is no 
' such tag.
Public Property Get rgszTag(iszTag)
	Call TraceEnterEx("Xnav.rgszTag", CStr(iszTag), tef)
	
	If iszTag >= 0 And iszTag < dictTagFunc.Count Then
		Dim rgsz : rgsz = dictTagFunc.Keys
		rgszTag = rgsz(iszTag)
	Else
		rgszTag = ""
	End If
	
	Call TraceExit("Xnav.rgszTag", SzQuote(rgszTag))
End Property


' Create a new waypoint of the type specified by szTag.
Public Function WayptFromTag(szTag)
	Call TraceEnterEx("Xnav.WayptFromTag", SzQuote(szTag), tef)
	
	Call Assert(dictTagFunc.Exists(szTag), _
	  "Unknown waypoint tag " + SzQuote(szTag))
	
	Set WayptFromTag = Eval(dictTagFunc(szTag) + "()")
	WayptFromTag.fWalk = False
	
	Call TraceExit("Xnav.WayptFromTag", SzTraceWaypt(WayptFromTag))
End Function

' Instantiate the given XML element as a waypoint of the appropriate type.
Public Function WayptFromXml(elem)
	Call TraceEnterEx("Xnav.WayptFromXml", SzTraceElem(elem), tef)
	
	Dim waypt : Set waypt = WayptFromTag(elem.tagName)
	waypt.fWalk = FGetAttr(elem, "fWalk", False)
	Call waypt.FromXml(elem)
	Set WayptFromXml = waypt
	
	Call TraceExit("Xnav.WayptFromXml", SzTraceWaypt(WayptFromXml))
End Function

' Save the given waypt as an XML element.  If fVerbose is True, all waypoint 
' parameters are made explicit in the XML, even those with default values.  If 
' fVerbose is False, parameters with default values may be omitted.
Public Function ElemFromWaypt(xdoc, waypt, fVerbose)
	Call TraceEnterEx("Xnav.ElemFromWaypt", _
	  SzTraceXdoc(xdoc) + ", " + SzTraceWaypt(waypt) + ", " + CStr(fVerbose), tef)
	
	Dim elem : Set elem = xdoc.CreateElement(waypt.szTag)
	Call waypt.ToXml(elem, fVerbose)
	If waypt.fWalk Or fVerbose Then
		Call SetFAttr(elem, "fWalk", waypt.fWalk)
	End If
	Set ElemFromWaypt = elem
	
	Call TraceExit("Xnav.ElemFromWaypt", SzTraceElem(ElemFromWaypt))
End Function

' Return True if the given waypoint is of the Simple type (with no special 
' arrival action), False otherwise.
Public Function FSimpleWaypt(waypt)
	FSimpleWaypt = (waypt.szTag = "Simple")
End Function

Public Function SzTraceWaypt(waypt)
	
	If waypt Is Nothing Then
		SzTraceWaypt = "Nothing"
	Else
		SzTraceWaypt = waypt.szTag + "(" + SzTraceMaploc(waypt.maplocIn) + ")"
	End If
	
End Function


' Decode map coordinates stored on the given XML element and return them in 
' maploc form.
Public Function MaplocOfElem(elem)
'	Call TraceEnter("Xnav.MaplocOfElem", SzTraceElem(elem))
	
	If IsNull(elem.GetAttribute("lat")) Then
		Set MaplocOfElem = Nothing
	Else
		Dim lat : lat = FloatGetAttr(elem, "lat", 0)
		Dim lng : lng = FloatGetAttr(elem, "lng", 0)
		Dim z : z = FloatGetAttr(elem, "z", 0)
		Dim head : head = FloatGetAttr(elem, "head", 0)
		Set MaplocOfElem = skapi.MaplocFromLatLng(lat, lng, z, head)
	End If
	
'	Call TraceExit("Xnav.MaplocOfElem", SzTraceMaploc(MaplocOfElem))
End Function

' Encode map coordinates from the given maploc onto the given XML element.
Public Function SetElemMaploc(elem, maploc, fZ, fHead)
	
	Call SetFloatAttr(elem, "lat", maploc.lat, 4)
	Call SetFloatAttr(elem, "lng", maploc.lng, 4)
	If fZ Then Call SetFloatAttr(elem, "z", maploc.z, 2)
	If fHead Then Call SetFloatAttr(elem, "head", maploc.head, 1)
	
End Function


' Return the 2D distance in map units to the given ACO.
Public Function DistToAco(aco)
	Call TraceEnterEx("Xnav.DistToAco", SzTraceAco(aco), tef)
	
	DistToAco = skapi.maplocCur.Dist2DToMaploc(aco.maploc)
	
	Call TraceExit("Xnav.DistToAco", FormatNumber(DistToAco, 3))
End Function

' Return the 2D distance in map units to the given maploc.
Public Function DistToMaploc(maploc)
	Call TraceEnterEx("Xnav.DistToMaploc", SzTraceMaploc(maploc), tef)
	
	DistToMaploc = skapi.maplocCur.Dist2DToMaploc(maploc)
	
	Call TraceExit("Xnav.DistToMaploc", FormatNumber(DistToMaploc, 3))
End Function

' Return the 3D distance in map units to the given maploc.
Public Function Dist3DToMaploc(maploc)
	Call TraceEnterEx("Xnav.Dist3DToMaploc", SzTraceMaploc(maploc), tef)
	
	Dist3DToMaploc = skapi.maplocCur.Dist3DToMaploc(maploc)
	
	Call TraceExit("Xnav.Dist3DToMaploc", FormatNumber(Dist3DToMaploc, 3))
End Function

' Return the compass heading to the given maploc, i.e. the direction you'd 
' have to face to be aimed at that maploc.
Public Function HeadToMaploc(maploc)
	Call TraceEnterEx("Xnav.HeadToMaploc", SzTraceMaploc(maploc), tef)
	
	HeadToMaploc = skapi.maplocCur.HeadToMaploc(maploc)
	
	Call TraceExit("Xnav.HeadToMaploc", FormatNumber(HeadToMaploc, 1))
End Function

' Return the difference in compass heading (the amount you would have to turn by) 
' from the current location to the given maploc.
Public Function DheadToMaploc(maploc)
	Call TraceEnterEx("Xnav.DheadToMaploc", SzTraceMaploc(maploc), tef)
	
	Dim maplocCurT : Set maplocCurT = skapi.maplocCur
	DheadToMaploc = DheadNormalize(maplocCurT.HeadToMaploc(maploc) - _
	  maplocCurT.head)
	
	Call TraceExit("Xnav.DheadToMaploc", FormatNumber(DheadToMaploc, 1))
End Function


' Normalize the given heading to within 0 to 360 degrees.
Public Function HeadNormalize(head)
	Call TraceEnterEx("Xnav.HeadNormalize", FormatNumber(head, 1), tef)
	
	Dim headT : headT = head
	Do While headT >= 360
		headT = headT - 360
	Loop
	Do While headT < 0
		headT = headT + 360
	Loop
	HeadNormalize = headT
	
	Call TraceExit("Xnav.HeadNormalize", FormatNumber(HeadNormalize, 1))
End Function

' Normalize the given turning angle to within plus-or-minus 180 degrees.
Public Function DheadNormalize(dhead)
	Call TraceEnterEx("Xnav.DheadNormalize", FormatNumber(dhead, 1), tef)
	
	Dim dheadT : dheadT = dhead
	Do While dheadT > 180
		dheadT = dheadT - 360
	Loop
	Do While dheadT < -180
		dheadT = dheadT + 360
	Loop
	DheadNormalize = dheadT
	
	Call TraceExit("Xnav.DheadNormalize", FormatNumber(DheadNormalize, 1))
End Function

' Return the given maploc with its heading reversed by 180 degrees.
Public Function MaplocReverseHead(maploc)
	Call TraceEnter("Xnav.MaplocReverseHead", SzTraceMaploc(maploc))
	
	Set MaplocReverseHead = MaplocAddDhead(maploc, 180)
	
	Call TraceExit("Xnav.MaplocReverseHead", SzTraceMaploc(MaplocReverseHead))
End Function

' Return the given maploc with its heading altered by dhead degrees.
Public Function MaplocAddDhead(maploc, dhead)
	Call TraceEnter("Xnav.MaplocAddDhead", _
	  SzTraceMaploc(maploc) + ", " + FormatNumber(dhead, 1))
	
	Set MaplocAddDhead = skapi.MaplocFromLatLng(maploc.lat, maploc.lng, _
	  maploc.z, Xnav.HeadNormalize(maploc.head + dhead))
	
	Call TraceExit("Xnav.MaplocAddDhead", SzTraceMaploc(MaplocAddDhead))
End Function

Public Function FIndoorMaploc(maploc)
	FIndoorMaploc = maploc.fIndoors
End Function


' Enter the given portal and wait until we arrive at the other end.
Public Function UseAcoPortal(aco)
	Call TraceEnterEx("Xnav.UseAcoPortal", SzTraceAco(aco), tef)
	
	Call Assert(FAcoIsPortal(aco), "not a portal")
	
	Call aco.Use()
	
	Call WaitForPortal()
	
	Call TraceExit("Xnav.UseAcoPortal", "")
End Function

Public Function WaitForPortal()
	Call TraceEnterEx("Xnav.WaitForPortal", "", tef)
	
'	Call skapi.WaitEvent(20000, wemSpecific, evidOnStartPortalSelf)
	mstate = mstateStop
	Call skapi.WaitEvent(60000, wemSpecific, evidOnEndPortalSelf)
	
	Call TraceExit("Xnav.WaitForPortal", "")
End Function


'------------------------------------
' Public for use by XrouteCls et al.
'------------------------------------

Public fStopNav
Public crouteInProgress

' Motion state:
Public Property Get mstateUnknown()
	mstateUnknown = -1
End Property
Public Property Get mstateStop()
	mstateStop = 0
End Property
Public Property Get mstateWalk()
	mstateWalk = 1
End Property
Public Property Get mstateRun()
	mstateRun = 2
End Property

Public Property Get mstate()
	
	If mstateMbr = mstateUnknown Then
		If Not fWalkForward Then
			mstateMbr = mstateStop
		ElseIf fHoldRun = fRunAsDefault Then
			mstateMbr = mstateWalk
		Else
			mstateMbr = mstateRun
		End If
	End If
	
	mstate = mstateMbr
	
End Property

' Set our state of motion (running, walking, or stopped) to mstateNew.
Public Property Let mstate(mstateNew)
	Call TraceEnterEx("Xnav.mstate", " <- " + CStr(mstateNew), tef)
	
	If mstateMbr <> mstateNew Then
		Select Case mstateNew
			
		Case mstateStop
			Call skapi.KeylessAutorun(False)
			Call skapi.Keys(cmidWalkForward, kmoUp)
			Call skapi.Keys(cmidHoldRun, kmoUp)
			
		Case mstateWalk
			If fRunAsDefault Then
				Call skapi.Keys(cmidHoldRun, kmoDown)
			Else
				Call skapi.Keys(cmidHoldRun, kmoUp)
			End If
			Call skapi.Keys(cmidWalkForward, kmoDown)
			
		Case mstateRun
			Call skapi.KeylessAutorun(True)
			Call skapi.Keys(cmidWalkForward, kmoUp)
			Call skapi.Keys(cmidHoldRun, kmoUp)
			
		End Select
		
		mstateMbr = mstateNew
	End If
	
	Call TraceExit("Xnav.mstate", "")
End Property

Public Function XrouteFromXdocSzName(xdoc, szName)
	Call TraceEnterEx("Xnav.XrouteFromXdocSzName", SzQuote(szName), tef)
	
	If xdoc Is xdocRecent Then
		If dictSzXrouteRecent.Exists(szName) Then
			Set XrouteFromXdocSzName = dictSzXrouteRecent(szName)
		Else
			Set XrouteFromXdocSzName = Nothing
		End If
	Else
		Set XrouteFromXdocSzName = Nothing
	End If
	
	Call TraceExit("Xnav.XrouteFromXdocSzName", SzTraceObj(XrouteFromXdocSzName))
End Function


'----------
' Privates
'----------

Private dictTagFunc

Private xdocRecent
Private dictSzXrouteRecent

Private mstateMbr

Private Property Get fRunAsDefault()
	fRunAsDefault = ((skapi.chop And chopRunAsDefault) <> 0)
End Property

Private Property Get fWalkForward()
	fWalkForward = (skapi.KeyState(cmidWalkForward) = 2)
End Property

Private Property Get fHoldRun()
	fHoldRun = (skapi.KeyState(cmidHoldRun) = 2)
End Property


End Class


' Instantiate the global object.
Public Xnav
Set Xnav = New XnavCls


