Option Explicit
Class XrouteCls


'------------
' Properties
'------------

Public szName			' Name of this route, if any.
Public cwaypt			' Number of waypoints on the route.
Public iwaypt			' Which waypoint to resume from.
Public xrouteRev		' Explicit reverse route; may be Nothing.

' Return the ith waypoint in the route.
Public Property Get rgwaypt(iwaypt)
	Set rgwaypt = rgwayptMbr(iwaypt)
End Property

' Return the starting coords of the route.
Public Property Get maplocFirst()
	
	If cwaypt > 0 Then
		Set maplocFirst = rgwayptMbr(0).maplocIn
	Else
		Set maplocFirst = Nothing
	End If
	
End Property

' Return the ending coords of the route.
Public Property Get maplocLast()
	
	If cwaypt > 0 Then
		Set maplocLast = rgwayptMbr(cwaypt - 1).maplocOut
	Else
		Set maplocLast = Nothing
	End If
	
End Property

' Estimate the end-to-end travel time of this route.
Public Property Get csecTravel()
	Call TraceEnterEx("Xroute.csecTravel", szTrace, Xnav.tef)
	
	csecTravel = 0
	
	Dim maplocPrev : Set maplocPrev = Nothing
	Dim vWalk : vWalk = Xnav.vWalk
	Dim vRun : vRun = skapi.vRun
	
	Dim iwayptT
	For iwayptT = 0 To cwaypt - 1
		Dim waypt : Set waypt = rgwayptMbr(iwayptT)
		If Not (maplocPrev Is Nothing) Then
			Dim v : If waypt.fWalk Then v = vWalk Else v = vRun
			csecTravel = csecTravel + maplocPrev.Dist2DToMaploc(waypt.maplocIn)/v
		End If
		csecTravel = csecTravel + waypt.csecArrive
		Set maplocPrev = waypt.maplocOut
	Next
	
	Call TraceExit("Xroute.csecTravel", FormatNumber(csecTravel, 1))
End Property

' Return a descriptive string suitable for use with TraceEnter/Exit.
Public Property Get szTrace()
	
	If szName <> "" Then
		szTrace = SzQuote(szName)
	ElseIf cwaypt > 0 Then
		szTrace = maplocLast.sz(3)
	Else
		szTrace = "empty"
	End If
	
End Property


'---------
' Methods
'---------

' Delete all waypoints from the route and reset it to an empty state.
Public Function Clear()
	Call TraceEnterEx("Xroute.Clear", szTrace, Xnav.tef)
	
	szName = ""
	Call Resize(0)
	iwaypt = 0
	Set xrouteRev = Nothing
	
	Call TraceExit("Xroute.Clear", "")
End Function

' Insert a normal waypoint (one with no arrival action) at position iwayptIns 
' in the route.  maploc specifies the map location of the waypoint.
Public Function InsertMaploc(iwayptIns, maploc)
	Call TraceEnterEx("Xroute.InsertMaploc", szTrace + ", " + _
	  CStr(iwayptIns) + ", " + SzTraceMaploc(maploc), Xnav.tef)
	
	Dim waypt : Set waypt = Xnav.WayptFromTag("Simple")
	Call waypt.Init(maploc, -1)
	Call InsertWaypt(iwayptIns, waypt)
	
	Call TraceExit("Xroute.InsertMaploc", "")
End Function

' Insert a waypoint of arbitrary type at position iwayptIns in the route.  
' Creating and initializing the waypt object is the caller's responsibility.
Public Function InsertWaypt(iwayptIns, waypt)
	Call TraceEnterEx("Xroute.InsertWaypt", szTrace + ", " + _
	  CStr(iwayptIns) + ", " + Xnav.SzTraceWaypt(waypt), Xnav.tef)
	
	If iwayptIns = iNil Then iwayptIns = cwaypt
	
	Call Resize(cwaypt + 1)
	
	Dim iwayptT
	For iwayptT = cwaypt - 1 To iwayptIns + 1 Step -1
		Set rgwayptMbr(iwayptT) = rgwayptMbr(iwayptT - 1)
	Next
	
	Set rgwayptMbr(iwayptIns) = waypt
	
	Call TraceExit("Xroute.InsertWaypt", "")
End Function

' Delete the waypoint at position iwayptDel in the route.
Public Function DeleteWaypt(iwayptDel)
	Call TraceEnterEx("Xroute.DeleteWaypt", szTrace + ", " + CStr(iwayptDel), Xnav.tef)
	
	Dim iwayptT
	For iwayptT = iwayptDel To cwaypt - 2
		Set rgwayptMbr(iwayptT) = rgwayptMbr(iwayptT + 1)
	Next
	
	Call Resize(cwaypt - 1)
	
	Call TraceExit("Xroute.DeleteWaypt", "")
End Function

' Append the points in xroute2 to the end of this route, concatenating the two 
' routes.  xroute2 is not modified by this method (but this route is).
Public Function Append(xroute2, fOptimize)
	Call TraceEnterEx("Xroute.Append", _
	  szTrace + ", " + SzTraceObj(xroute2) + ", " + CStr(fOptimize), Xnav.tef)
	
	Dim iwayptFirst1 : iwayptFirst1 = 0
	Dim iwayptLim1 : iwayptLim1 = cwaypt
	Dim iwayptFirst2 : iwayptFirst2 = 0
	Dim iwayptLim2 : iwayptLim2 = xroute2.cwaypt
'	Call TraceLine("iwayptLim1 = " + CStr(iwayptLim1) + ", " + _
'	  "iwayptFirst2 = " + CStr(iwayptFirst2))
	
	If fOptimize Then
	'	Try to prune away any pointless backtracking from the juncture of the two 
	'	routes.
		Do While iwayptLim1 > iwayptFirst1 And iwayptFirst2 < iwayptLim2
			Dim waypt1 : Set waypt1 = rgwayptMbr(iwayptLim1 - 1)
			Dim waypt2 : Set waypt2 = xroute2.rgwaypt(iwayptFirst2)
			
		'	If the last point of xroute1 coincides with the first point of xroute2, 
		'	omit them both.
			If waypt1.maplocOut.Dist2DToMaploc(waypt2.maplocIn) > 2 * Xnav.distMeter Then Exit Do
			If Not Xnav.FSimpleWaypt(waypt1) And Not Xnav.FSimpleWaypt(waypt2) Then Exit Do
			
			If Xnav.FSimpleWaypt(waypt1) Then iwayptLim1 = iwayptLim1 - 1
			If Xnav.FSimpleWaypt(waypt2) Then iwayptFirst2 = iwayptFirst2 + 1
		'	Call TraceLine("iwayptLim1 = " + CStr(iwayptLim1) + ", " + _
		'	  "iwayptFirst2 = " + CStr(iwayptFirst2))
		Loop
	End If
	
	Call Resize(iwayptLim1 + iwayptLim2 - iwayptFirst2)
	
	Dim iwaypt2
	For iwaypt2 = iwayptFirst2 To iwayptLim2 - 1
		If xroute2.rgwaypt(iwaypt2).maplocIn Is Nothing Then
		'	Found a recall or something that can happen anywhere.  No point 
		'	hanging onto anything that comes before it.
			iwayptLim1 = 0
		End If
	'	Call TraceLine("iwayptLim1 = " + CStr(iwayptLim1) + ", " + _
	'	  "iwaypt2 = " + CStr(iwaypt2))
		Set rgwayptMbr(iwayptLim1) = xroute2.rgwaypt(iwaypt2)
		iwayptLim1 = iwayptLim1 + 1
	Next
	
'	Call TraceLine("iwayptLim1 = " + CStr(iwayptLim1))
	Call Resize(iwayptLim1)
	
	Call TraceExit("Xroute.Append", "")
End Function

' Return a new Xroute representing the concatenation of this route with 
' xroute2.  Neither this route nor xroute2 is modified by this method.
Public Function XrouteConcat(xroute2, fOptimize)
	Call TraceEnterEx("Xroute.XrouteConcat", _
	  szTrace + ", " + SzTraceObj(xroute2) + ", " + CStr(fOptimize), Xnav.tef)
	
	Dim xrouteNew : Set xrouteNew = Xnav.XrouteNew()
	Call xrouteNew.Append(Me, False)
	Call xrouteNew.Append(xroute2, fOptimize)
	
	Set XrouteConcat = xrouteNew
	
	Call TraceExit("Xroute.XrouteConcat", SzTraceObj(XrouteConcat))
End Function

' Return a new Xroute that is an exact copy of this route.
Public Function XrouteClone()
	Call TraceEnterEx("Xroute.XrouteClone", szTrace, Xnav.tef)
	
	Dim xrouteNew : Set xrouteNew = Xnav.XrouteNew()
	xrouteNew.szName = szName
	Call xrouteNew.Append(Me, False)
	
	Set XrouteClone = xrouteNew
	
	Call TraceExit("Xroute.XrouteClone", SzTraceObj(XrouteClone))
End Function

' Return a new Xroute with with order of the points reversed.
Public Function XrouteReverse()
	Call TraceEnterEx("Xroute.XrouteReverse", szTrace, Xnav.tef)
	
'	If there's an explicit reverse route, return that; otherwise calculate 
'	the literal point-by-point reversal.
	If xrouteRev Is Nothing Then
		Dim xrouteNew : Set xrouteNew = Xnav.XrouteNew()
		xrouteNew.szName = "~" + szName
		Set XrouteReverse = xrouteNew
		Set xrouteNew.xrouteRev = Me
		Dim fWalk : fWalk = False
		Dim iwayptT
		For iwayptT = cwaypt - 1 To 0 Step -1
		'	Ask each waypoint if it's reversible.
			Dim waypt : Set waypt = rgwayptMbr(iwayptT)
			Dim wayptRev : Set wayptRev = waypt.WayptReverse()
			If wayptRev Is Nothing Then
				Set XrouteReverse = Nothing
				Exit For
			End If
			wayptRev.fWalk = fWalk
			Call xrouteNew.InsertWaypt(iNil, wayptRev)
			fWalk = waypt.fWalk
		Next
	Else
		Set XrouteReverse = xrouteRev
	End If
	
	Call TraceExit("Xroute.XrouteReverse", SzTraceObj(XrouteReverse))
End Function


' Load route information from the Xroute file named by szFile.
Public Function FromFile(szFile, fInterpolate)
	Call TraceEnter("Xroute.FromFile", _
	  szTrace + ", " + SzQuote(szFile))
	
	Dim rex : Set rex = New RegExp
	rex.Pattern = "([^\[]+)(\[([^\]]+)\])?"
	rex.IgnoreCase = True
	rex.Global = False
	
	Dim szFileT, szRoute
	szFileT = rex.Replace(szFile, "$1")
	szRoute = rex.Replace(szFile, "$3")
	Call TraceLine("szFile = " + SzQuote(szFileT))
	Call TraceLine("szRoute = " + SzQuote(szRoute))
	
	Dim xdoc : Set xdoc = XdocLoadFile(szFileT)
	
	Dim elemXroute
	If szRoute <> "" Then
		Set elemXroute = xdoc.SelectSingleNode("//Xroute[@szName = " + _
		  SzQuote(szRoute) + "]")
	Else
		Set elemXroute = xdoc.SelectSingleNode("//Xroute")
	End If
	Call FromXml(elemXroute, fInterpolate)
	
	Call TraceExit("Xroute.FromFile", "")
End Function

' Load route information from the given XML text.
Public Function FromSz(szXml, fInterpolate)
	Call TraceEnter("Xroute.FromSz", szTrace)
	
	Dim xdoc : Set xdoc = XdocNew("")
	
	Call xdoc.LoadXml(szXml)
	
	Dim elemXroute : Set elemXroute = xdoc.SelectSingleNode("//Xroute")
	Call FromXml(elemXroute, fInterpolate)
	
	Call TraceExit("Xroute.FromSz", "")
End Function

' Load route information from the given XML element.
Public Function FromXml(elem, fInterpolate)
	Call TraceEnter("Xroute.FromXml", szTrace)
	
	Call Clear()
	
	szName = SzGetAttr(elem, "szName")
	
	Dim sel : Set sel = elem.SelectNodes("*")
	Dim elemWaypt
	For Each elemWaypt In sel
		Dim waypt : Set waypt = Xnav.WayptFromXml(elemWaypt)
		If fInterpolate And cwaypt > 0 Then
		'	Break up long legs by interpolating intermediate waypoints.
			Dim maplocPrev : Set maplocPrev = maplocLast
			Dim maplocNext : Set maplocNext = waypt.maplocIn
			Dim dist : dist = maplocPrev.Dist2DToMaploc(maplocNext)
			Dim distLeg : distLeg = 0.5
			Dim cleg : cleg = Round(dist/distLeg)
			Dim ileg
			For ileg = 1 To cleg - 1
				Dim fract : fract = ileg/cleg
				Dim maplocT : Set maplocT = skapi.MaplocFromLatLng( _
				  maplocPrev.lat * (1 - fract) + maplocNext.lat * fract, _
				  maplocPrev.lng * (1 - fract) + maplocNext.lng * fract)
				Dim wayptT : Set wayptT = Xnav.WayptFromTag("Simple")
				wayptT.fWalk = waypt.fWalk
				Call wayptT.Init(maplocT, -1)
				Call InsertWaypt(iNil, wayptT)
			Next
		End If
		Call InsertWaypt(iNil, waypt)
	Next
	
	Call TraceExit("Xroute.FromXml", "")
End Function

' Read an AC Explorer format route file.
Public Function FromAcxRte(szFile)
	Call TraceEnter("Xroute.FromAcxRte", _
	  szTrace + ", " + SzQuote(szFile))
	
	Dim ts : Set ts = fso.OpenTextFile(szFile, ForReading)
	
	Call Clear()
	
	Do Until ts.AtEndOfStream
		Dim szLine : szLine = ts.ReadLine()
		Call TraceLine(szLine)
		
		If FPrefixMatchSz(szLine, "!") Then
			Dim szKind : szKind = SzExtractField(szLine, 2)
			Dim szMaploc, maploc
			szMaploc = SzExtractField(szLine, 3) + " " + SzExtractField(szLine, 4)
			Set maploc = skapi.MaplocFromSz(szMaploc)
			Select Case szKind
			Case "Portal"
			'	REVIEW: Portal name in AC-Spedia DB may not be accurate.
				Dim szPortal : szPortal = SzExtractField(szLine, 1)
				Dim maplocOut : Set maplocOut = skapi.MaplocFromSz( _
				  SzExtractField(szLine, 5) + " " + SzExtractField(szLine, 6))
				Dim waypt : Set waypt = Xnav.WayptFromTag("Portal")
				Call waypt.Init(szPortal, maploc, maplocOut)
				Call InsertWaypt(iNil, waypt)
			Case "TempLine"
				Call InsertMaploc(maploc)
				Dim szHead, szTail
				Call SplitSz(szHead, szTail, szLine, "=")
				If szTail <> "" Then
					Do While szTail <> ""
						Call SplitSz(szMaploc, szTail, szTail, " ")
						Call InsertMaploc(iNil, skapi.MaplocFromSz(szMaploc))
					Loop
				Else
					szMaploc = SzExtractField(szLine, 5) + "," + _
					  SzExtractField(szLine, 6)
					Call InsertMaploc(iNil, skapi.MaplocFromSz(szMaploc))
				End If
			Case Else
				Call InsertMaploc(iNil, maploc)
			End Select
		End If
	Loop
	
	Call ts.Close()
	
	Call TraceExit("Xroute.FromAcxRte", "")
End Function


' Save the route to the file named by szFile.
Public Function SaveToFile(szFile)
	Call TraceEnterEx("Xroute.SaveToFile", _
	  szTrace + ", " + SzQuote(szFile), Xnav.tef)
	
	Call XdocSave().Save(szFile)
	
	Call TraceExit("Xroute.SaveToFile", "")
End Function

' Save the route to a new XML document in memory.
Public Function XdocSave()
	Call TraceEnterEx("Xroute.XdocSave", szTrace, Xnav.tef)
	
	Dim xdoc : Set xdoc = Xnav.XdocNewXrouteFile()
	Call xdoc.documentElement.AppendChild(ElemSave(xdoc))
	Call AdjustElemIndent(xdoc.documentElement, 0)
	
	Set XdocSave = xdoc
	
	Call TraceExit("Xroute.XdocSave", "")
End Function

' Create and returns an Xroute XML element representing the route.
Public Function ElemSave(xdoc)
	Call TraceEnterEx("Xroute.ElemSave", szTrace, Xnav.tef)
	
	Dim elemXroute : Set elemXroute = xdoc.CreateElement("Xroute")
	
	Call SetSzAttr(elemXroute, "szName", szName)
	
	Dim iwayptT
	For iwayptT = 0 To cwaypt - 1
		Dim waypt : Set waypt = rgwayptMbr(iwayptT)
		Dim elemWaypt : Set elemWaypt = Xnav.ElemFromWaypt(xdoc, waypt, False)
		If waypt.fWalk Then Call SetFAttr(elemWaypt, "fWalk", True)
		Call elemXroute.AppendChild(elemWaypt)
	Next
	
	Set ElemSave = elemXroute
	
	Call TraceExit("Xroute.ElemSave", "")
End Function


' Debugging aid: display the route definition in the chat window.
Public Function Px()
	Call TraceEnterEx("Xroute.Px", szTrace, Xnav.tef)
	
	Call skapi.OutputSz(XdocSave().xml, opmChatWnd)
	
	Call TraceExit("Xroute.Px", "")
End Function


' Move your character from waypoint to waypoint along the route, performing 
' each waypoint's arrival action as you come to it.
Public Function Go()
	Call TraceEnterEx("Xroute.Go", szTrace, Xnav.tef)
	
	iwaypt = 0
	Go = ResumeRoute()
	
	Call TraceExit("Xroute.Go", CStr(Go))
End Function

' Resume execution of an interrupted route.
Public Function ResumeRoute()
	Call TraceEnterEx("Xroute.ResumeRoute", szTrace, Xnav.tef)
	
	ResumeRoute = nscArrived
	
	Pauser.fStopOnCtrlBreak = True
	Pauser.fStopOnLogout = True
	Pauser.fStopOnDeath = True
	
'	Keep retrying until we reach the end of the route.
'	This takes care of pauses in mid-route.
	Xnav.fStopNav = False
	Do While iwaypt < cwaypt
		
	'	Error-handling boilerplate.
		Dim erh : Set erh = ErhBegin()
		On Error Resume Next
		
		Call ResumeEx()
		
	'	Error-handling boilerplate.
		Call EndErh(erh)
		On Error Goto 0
		If erh.errno <> errnoOK Then Call erh.PassItOn()
		
		Call Pauser.Check()
		If Xnav.fStopNav Then
			ResumeRoute = nscCanceled
			Exit Do
		End If
	Loop
	
	Call TraceExit("Xroute.ResumeRoute", CStr(ResumeRoute))
End Function


'----------------
' Event handlers
'----------------

Public Function OnNavStop(nsc)
	Call TraceEnterEx("Xroute.OnNavStop", CStr(nsc), Xnav.tef)
	Do
	
	nscBreak = nsc
	If nsc <> nscArrived Then
		fBreak = True
		Exit Do
	End If
	
	If iwaypt >= cwaypt Then Exit Do
	Dim waypt : Set waypt = rgwayptMbr(iwaypt)
	If Not Xnav.FSimpleWaypt(waypt) Then
	'	This waypoint requires special action before continuing.
		Xnav.mstate = Xnav.mstateStop		' REVIEW
		fBreak = True
		Exit Do
	End If
	
'	Step to the next waypoint, if any.
	iwaypt = iwaypt + 1
	If iwaypt = cwaypt Then
	'	We've arrived at the destination.
		Xnav.mstate = Xnav.mstateStop
		fBreak = True
		Exit Do
	End If
	
	Set waypt = rgwayptMbr(iwaypt)
	If waypt.maplocIn Is Nothing Then
	'	No place to go.
		Xnav.mstate = Xnav.mstateStop		' REVIEW
		fBreak = True
		Exit Do
	End If
	
	Call TraceLine("skapi.GoToMaploc(" + SzTraceMaploc(waypt.maplocIn) + ", " + _
	  FormatNumber(waypt.distArrive, 3) + ", " + CStr(waypt.fWalk) + ")")
	Call skapi.GoToMaploc(waypt.maplocIn, waypt.distArrive, waypt.fWalk, False)
	Xnav.mstate = Xnav.mstateUnknown
	
	Loop While False
	Call TraceExit("Xroute.OnNavStop", "")
End Function

Public Function OnStartPortalSelf()
	Call TraceEnterEx("Xroute.OnStartPortalSelf", "", Xnav.tef)
	
	If Not fPortalOK Then
		Call Pauser.SetError("Unexpected portal travel.")
	End If
	
	Call TraceExit("Xroute.OnStartPortalSelf", "")
End Function


'----------------
' Unwind handler
'----------------

Public Function Unwind(erh)
	Call TraceEnterEx("Xroute.Unwind", "", tefEnable)
	
	Xnav.crouteInProgress = Xnav.crouteInProgress - 1
	Call skapi.CancelNav()
	Xnav.mstate = Xnav.mstateStop
	
	Call skapi.RemoveHandler(evidOnNavStop, Me)
	Call skapi.RemoveHandler(evidOnStartPortalSelf, Me)
	
	Call TraceExit("Xroute.Unwind", "")
End Function


'----------
' Privates
'----------

Private rgwayptMbr()		' Array of waypoints on the route.
Private cwayptMal			' Actual allocated size of rgwayptMbr; may be different than cwaypt.
Private fBreak				' True if ResumeEx should break out of its inner loop. 
Private nscBreak			' The reason ResumeEx should break out of its inner loop.
Private fPortalOK			' True if portal travel is expected, False otherwise.

Private Sub Class_Initialize()
	
	szName = ""
	cwayptMal = 0
	cwaypt = 0
	iwaypt = 0
	Set xrouteRev = Nothing
	
	
End Sub

Private Function Resize(cwayptNew)
'	Call TraceEnter("Xroute.Resize", CStr(cwayptNew))
	
	If cwayptNew = 0 Then
		ReDim rgwayptMbr(0)
		cwayptMal = 1
		cwaypt = 0
	ElseIf cwayptNew > cwayptMal Then
		cwayptMal = cwayptNew + 20
		ReDim Preserve rgwayptMbr(cwayptMal - 1)
	'	Call TraceLine("cwayptMal = " + CStr(cwayptMal))
	End If
	
	Do While cwaypt > cwayptNew
		cwaypt = cwaypt - 1
		Set rgwayptMbr(cwaypt) = Nothing
	Loop
	
	cwaypt = cwayptNew
	
'	Call TraceExit("Xroute.Resize", "")
End Function


Private Function ResumeEx()
	Call TraceEnter("Xroute.ResumeEx", szTrace)
	Call PushUnwind(Me)
	
	Xnav.crouteInProgress = Xnav.crouteInProgress + 1
	Xnav.mstate = Xnav.mstateUnknown
	
	fPortalOK = False
	Call skapi.AddHandler(evidOnStartPortalSelf, Me)
	
	Dim iwayptStuck : iwayptStuck = iNil
	Dim itryStuck : itryStuck = 0
	
	Do While iwaypt < cwaypt
		Dim waypt : Set waypt = rgwayptMbr(iwaypt)
		If Not Pauser.fRunning Or Xnav.fStopNav Then Exit Do
		
		Call skapi.WaitEvent()
		
		If waypt.maplocIn Is Nothing Then
			nscBreak = nscArrived
		Else
			Call skapi.AddHandler(evidOnNavStop, Me)
			
			Call TraceLine("skapi.GoToMaploc(" + SzTraceMaploc(waypt.maplocIn) + ", " + _
			  FormatNumber(waypt.distArrive, 3) + ", " + CStr(waypt.fWalk) + ")")
			Call skapi.GoToMaploc(waypt.maplocIn, waypt.distArrive, waypt.fWalk, False)
			Xnav.mstate = Xnav.mstateUnknown
			
			fBreak = False
			Do Until fBreak
				If Not Pauser.fRunning Or Xnav.fStopNav Then
					nscBreak = nscCanceled
					Exit Do
				End If
				Call skapi.WaitEvent(500, wemSpecific, evidOnNavStop)
			Loop
			
			Call skapi.RemoveHandler(evidOnNavStop, Me)
		End If
		
		Select Case nscBreak
			
		Case nscArrived
			If iwaypt < cwaypt Then
			'	This waypoint needs special action such as opening a door, 
			'	jumping off a ledge, using a portal, etc.  Do it now.
				Set waypt = rgwayptMbr(iwaypt)
				fPortalOK = True
				Call waypt.Arrive()
				fPortalOK = False
				iwaypt = iwaypt + 1
			End If
			
		Case nscTimeout
		'	Failed to arrive at some waypoint.	We must be hung up on something.
		'	Back off and retry.
			If iwaypt = iwayptStuck Then
				itryStuck = itryStuck + 1
			Else
				iwayptStuck = iwaypt
				itryStuck = 0
			End if
			Call GetUnstuck(itryStuck)
			
		Case nscCanceled
		'	User interrupt.
			Exit Do
			
		End Select
	Loop
	
	Xnav.crouteInProgress = Xnav.crouteInProgress - 1
	If Xnav.crouteInProgress = 0 Then
		Call skapi.CancelNav()
		Xnav.mstate = Xnav.mstateStop
	End If
	
	Call skapi.RemoveHandler(evidOnStartPortalSelf, Me)
	
	Call PopUnwind(Me)
	Call TraceExit("Xroute.ResumeEx", "")
End Function

Private Function GetUnstuck(itry)
	Call TraceEnter("Xroute.GetUnstuck", CStr(itry))
	
	Xnav.mstate = Xnav.mstateStop
	
	If iwaypt > 0 And itry < 1 Then
	'	Head back toward previous waypoint.
		Dim cmsec : cmsec = (itry + 1) * 5000
		Call skapi.GoToMaploc(rgwayptMbr(iwaypt - 1).maplocOut, Xnav.distMeter, False, True, cmsec)
		Xnav.mstate = Xnav.mstateUnknown
		Call skapi.WaitEvent(cmsec, wemSpecific, evidOnNavStop)
	ElseIf itry < 2 Then
	'	Head roughly back the way we came, but with a bit of randomness.
 		Dim head : head = Xnav.HeadNormalize(skapi.maplocCur.head + 135 + Rnd() * 90)
		Call Xnav.TurnToHead(head)
		
		Xnav.mstate = Xnav.mstateRun
		Call skapi.WaitEvent(5000, wemFullTimeout)
		Xnav.mstate = Xnav.mstateStop
	Else
		Call RaiseCmr("Could not complete route.")
	End If
	
	Call TraceExit("Xroute.GetUnstuck", "")
End Function


End Class


' Instantiator function.
Public Function XrouteNew()
	Set XrouteNew = New XrouteCls
End Function


