Option Explicit
Class XrouteConvertCls

' SkunkNav-to-Xnav route converter.


'-------------------
' Class boilerplate
'-------------------

Public tef

Private Sub Class_Initialize()
	Call TraceEnter("XrouteConvert.Initialize", "")
	
	Call Trace(Me)
	
	Call TraceExit("XrouteConvert.Initialize", "")
End Sub

Private Sub Class_Terminate()
	Call TraceEnter("XrouteConvert.Terminate", "")
	
	Call TraceExit("XrouteConvert.Terminate", "")
End Sub


'---------
' Methods
'---------

' Search the specified folder for SkunkNav-format .route files and convert 
' them to Xnav-format .xroute files.  The original .route files are not 
' altered or deleted, but new .xroute files are created alongside them.
Public Function ConvertFolder(szFolder)
	Call TraceEnter("XrouteConvert.ConvertFolder", SzQuote(szFolder))
	
	Dim fldr : fldr = fso.GetFolder(szFolder)
	Dim cofil : Set cofil = fldr.Files
	Dim fil
	For Each fil In cofil
		If FSuffixMatchSz(fil.Name, ".route") Then
			ConvertFile(fil.Path)
		End If
		Call Pauser.Check()
	Next
	
	Call TraceExit("XrouteConvert.ConvertFolder", "")
End Function

' Convert the specified SkunkNav-format .route file to an Xnav-format .xroute 
' file.  The original .route file is not altered or deleted, but a new .xroute 
' file is created alongside it. 
Public Function ConvertFile(szFile)
	Call TraceEnter("XrouteConvert.ConvertFile", SzQuote(szFile))
	
	Call skapi.OutputLine("Converting " + fso.GetFileName(szFile) + "...")
	
	Dim xdoc : Set xdoc = XdocReadFile(szFile)
	Call AdjustElemIndent(xdoc.documentElement, 0)
	Call xdoc.Save(fso.BuildPath(fso.GetParentFolderName(szFile), _
	  fso.GetBaseName(szFile) + ".xroute"))
	
	Call TraceExit("XrouteConvert.ConvertFile", "")
End Function

' Read the specified SkunkNav-format .route file into memory as an Xnav-format 
' XML document and returns a reference to that document.  The original .route 
' file on disk is not altered or deleted.
Public Function XdocReadFile(szFile)
	Call TraceEnter("XrouteConvert.XdocReadFile", SzQuote(szFile))
	
	Dim xdoc : Set xdoc = Xnav.XdocNewXrouteFile()
	
	Dim ts : Set ts = fso.OpenTextFile(szFile, ForReading)
	Dim iroute : iroute = 0
	Do Until ts.AtEndOfStream
		Call ReadRoute(xdoc, ts, iroute)
		iroute = iroute + 1
	Loop
	Call ts.Close()
	
	Set XdocReadFile = xdoc
	
	Call TraceExit("XrouteConvert.XdocReadFile", SzTraceXdoc(XdocReadFile))
End Function


'----------
' Privates
'----------

Private Property Get distWalk()
	distWalk = 0.035
End Property
Private Property Get distArriveMin()
	distArriveMin = 0.0075
End Property

Private Function ReadRoute(xdoc, ts, iroute)
	Call TraceEnter("XrouteConvert.ReadRoute", "")
	
	Dim xroute : Set xroute = Xnav.XrouteNew()
	
'	Assign a default route name in case the route doesn't specify one.
	If iroute = 0 Then
		xroute.szName = "Default route"
	Else
		xroute.szName = "Route " + CStr(iroute + 1)
	End If
	
	Dim wayptPrev : Set wayptPrev = Nothing

	Dim fBlank : fBlank = True
	Do Until ts.AtEndOfStream
		Dim szLine : szLine = ts.ReadLine()
		Dim waypt : Set waypt = Nothing
		
		If szLine = "" Or FPrefixMatchSz(szLine, "//") Then
		'	Blank/comment line; ignore.
		
		ElseIf FPrefixMatchSz(szLine, "[") Then
		'	Name of this route.
			xroute.szName = SzStripBrackets(szLine, "[", "]")
			fBlank = False
		
		ElseIf FPrefixMatchSz(szLine, "@") Then
		'	Recall command.
			Set waypt = Xnav.WayptFromTag("Recall")
		'	Destination coords are unknown.
			Call waypt.Init(szLine, Nothing, skapi.MaplocFromLatLng(0, 0))
			
		ElseIf FEqSzI(szLine, "Reverse") Then
		'	Explicit reverse route (to handle portal/recall case).
			Call xdoc.documentElement.AppendChild(xroute.ElemSave(xdoc))
			Dim szNameRev : szNameRev = "~" + xroute.szName
			Set xroute = Xnav.XrouteNew()
			xroute.szName = szNameRev
			Set wayptPrev = Nothing
			fBlank = False
		
		ElseIf FEqSzI(szLine, "Stop") Then
		'	End of route.
			fBlank = False
			Exit Do
		
	'	Use semicolon as value separator to avoid localization confusion 
	'	in locales where comma is the decimal separator.
		ElseIf FContainsSz(szLine, ";") Then
		'	Waypoint is an object; decode location and name.
			Set waypt = WayptFromSzAco(szLine, ";")
			
	'	Accept comma as value separator for compatibility, 
	'	but only if there are periods in the line as well.
		ElseIf FContainsSz(szLine, ",") And _
		  FContainsSz(szLine, ".") Then
		'	Waypoint is an object; decode location and name.
			Set waypt = WayptFromSzAco(szLine, ",")
			
		Else
		'	Ordinary waypoint.
			Dim maploc : Set maploc = skapi.MaplocFromSz(szLine)
			Dim distArrive
			If wayptPrev Is Nothing Then
				distArrive = distWalk
			ElseIf wayptPrev.maplocOut.Dist2DToMaploc(maploc) >= distWalk Then
				distArrive = distWalk
			Else
				distArrive = distArriveMin
				If Xnav.FSimpleWaypt(wayptPrev) Then
					Call wayptPrev.Init(wayptPrev.maplocIn, distArrive)
				End If
			End If
			Set waypt = Xnav.WayptFromTag("Simple")
			Call waypt.Init(maploc, distArrive)
			
		End If
		
		If Not (waypt Is Nothing) Then
			If Not (wayptPrev Is Nothing) And Not (waypt.maplocIn Is Nothing) Then
				If wayptPrev.maplocOut.Dist2DToMaploc(waypt.maplocIn) < distWalk Then
					waypt.fWalk = True
				End If
			End If
			Call xroute.InsertWaypt(iNil, waypt)
			Set wayptPrev = waypt
			fBlank = False
		End If
	Loop
	
	If Not fBlank Then
		Call xdoc.documentElement.AppendChild(xroute.ElemSave(xdoc))
	End If
	
	Call TraceExit("XrouteConvert.ReadRoute", "")
End Function

Private Function WayptFromSzAco(szLine, szSep)
	Call TraceEnter("XrouteConvert.WayptFromSzAco", SzQuote(szLine) + ", " + SzQuote(szSep))
	
	Dim szMaploc, szAco, szTail
	Call SplitSz(szMaploc, szTail, szLine, szSep)
	Call SplitSz(szAco, szTail, szTail, szSep)
	Call TraceLine("szMaploc = " + SzQuote(szMaploc) + ", " + _
	  "szAco = " + SzQuote(szAco))
	
	Dim maploc : Set maploc = skapi.MaplocFromSz(szMaploc)
	Dim waypt
	If FSuffixMatchSz(szAco, "Door") Then
		Set waypt = Xnav.WayptFromTag("Door")
	'	Door's heading is unknown.
		Call waypt.Init(skapi.MaplocFromLatLng(maploc.lat, maploc.lng, 0, -1), "")
	Else
		Set waypt = Xnav.WayptFromTag("Portal")
	'	Portal destination coords are unknown.
		Call waypt.Init(szAco, maploc, skapi.MaplocFromLatLng(0, 0))
	End If
	
	Set WayptFromSzAco = waypt
	
	Call TraceExit("XrouteConvert.WayptFromSzAco", Xnav.SzTraceWaypt(WayptFromSzAco))
End Function


End Class


' Instantiate the global object.
Public XrouteConvert
Set XrouteConvert = New XrouteConvertCls


