Option Explicit
Class XrouteEditCls

' This module implements an editor for the XML route files used by Xnav.  
' The sole public method is XrouteEdit.Edit, which displays the in-game 
' route-editing controls.  Editing features are similar to the RouteMaker 
' script supplied with SkunkNav, adapted to accommodate the extensible 
' waypoint capabilities of Xnav.  There's also a new Record feature that 
' automatically records your movements as you navigate around manually.


'-------------------
' Class boilerplate
'-------------------

Public tef

Private Sub Class_Initialize()
	Call TraceEnter("XrouteEdit.Initialize", "")
	
	Call Trace(Me)
	
	Call Reset()
	
	Call skapi.AddHandler(evidOnStart3D, Me)
	
	Call TraceExit("XrouteEdit.Initialize", "")
End Sub

Private Sub Class_Terminate()
	Call TraceEnter("XrouteEdit.Terminate", "")
	
	Call TraceExit("XrouteEdit.Terminate", "")
End Sub


'------------
' Properties
'------------

' Return True if the editing panel is displayed, False otherwise.
Public Property Get fDisplayed()
	fDisplayed = (szPanelName <> "")
End Property


'---------
' Methods
'---------

Public Function Edit(szFile)
	Call TraceEnterEx("XrouteEdit.Edit", SzQuote(szFile), tef)
	
	If szPanelName = "" Then
	'	Load and show the editor panel.
		Dim xdocControls : Set xdocControls = XdocLoadFile("XrouteEdit.xml")
		szPanelName = xdocControls.documentElement.GetAttribute("title")
		Call skapi.ShowControls(xdocControls.xml, True)
		
		Set dictSzFEnabled = CreateObject("Scripting.Dictionary")
		dictSzFEnabled.CompareMode = vbTextCompare
		dictSzFEnabled("lboxRoute") = True
		dictSzFEnabled("listWaypts") = True
		dictSzFEnabled("lboxAttr") = True
		dictSzFEnabled("btnNewFile") = True
		dictSzFEnabled("btnOpen") = True
		dictSzFEnabled("btnSave") = True
		dictSzFEnabled("btnRevert") = True
		dictSzFEnabled("btnQuit") = True
		
	'	Populate the waypoint type dropdown.
		Dim iszSel : iszSel = -1
		Dim iszTag, szTag
		Do
			szTag = Xnav.rgszTag(iszTag)
			If szTag = "" Then Exit Do
			Call skapi.GetControlProperty(szPanelName, "lboxTag", _
			  "AddChoice(" + SzQuote(szTag) + ")")
			If szTag = "Simple" Then iszSel = iszTag
			iszTag = iszTag + 1
		Loop
		Call skapi.SetControlProperty(szPanelName, "lboxTag", "Selected", CLng(iszSel))
		
	'	Prepare to handle control events.
		Call skapi.AddHandler(evidOnControlEvent, Me)
	End If
	
'	Load the specified route file.
	If szFile <> "" Then
		Call LoadRouteFile(fso.GetAbsolutePathName(szFile))
	Else
		Call NewRouteFile()
	End If
	
	Call TraceExit("XrouteEdit.Edit", "")
End Function


'----------------
' Event handlers
'----------------

Public Function OnStart3D()
	Call TraceEnterEx("XrouteEdit.OnStart3D", "", tef)
	
	Call Reset()
	
	Call TraceExit("XrouteEdit.OnStart3D", "")
End Function

Public Function OnControlEvent(szPanel, szControl, dictSzValue)
	Call TraceEnterEx("XrouteEdit.OnControlEvent", SzQuote(szControl), tef)
	
	If szPanel = szPanelName Then
	'	Error-handling boilerplate.
		Dim erh : Set erh = ErhBegin()
		On Error Resume Next
		
		Call DispatchControlEvent(szControl, dictSzValue)
		
	'	Error-handling boilerplate.
		Call EndErh(erh)
		On Error Goto 0
		
		If erh.errno <> errnoOK Then
			Call skapi.OutputLine(erh.sz, opmChatWnd, cmcYellow)
		End If
	End If
	
	Call TraceExit("XrouteEdit.OnControlEvent", "")
End Function

Public Function OnControlProperty(szPanel, szControl, szProperty, vValue)
	Call TraceEnterEx("XrouteEdit.OnControlProperty", _
	  SzQuote(szControl) + ", " + SzQuote(szProperty), tef)
	
	If FEqSzI(szPanel, szPanelName) Then
		Select Case szControl
			
		Case "listWaypts"
			Dim fWalk : fWalk = CBool(vValue)
			xrouteCur.rgwaypt(iwayptSetWalk).fWalk = fWalk
			fDirty = True
			Call skapi.RemoveHandler(evidOnControlProperty, Me)
			
		End Select
	End If
	
	Call TraceExit("XrouteEdit.OnControlProperty", "")
End Function

Public Function OnLocChangeSelf(maploc)
	Call TraceEnterEx("XrouteEdit.OnLocChangeSelf", SzTraceMaploc(maploc), tef Or tefNonEmpty)
	
	Select Case state
	Case statePlayback
		If iwayptSel <> xrouteCur.iwaypt Then
			If iwayptSel >= 0 And iwayptSel < xrouteCur.cwaypt Then
			'	Refresh waypoint as we leave it in case its parameters got updated.
				Call ShowWaypt(iwayptSel)
			End If
			iwayptSel = xrouteCur.iwaypt
			Call ScrollToWaypt(iwayptSel)
		End If
	End Select
	
	Call TraceExit("XrouteEdit.OnLocChangeSelf", "")
End Function

Public Function OnTimer(timer)
	Call TraceEnterEx("XrouteEdit.OnTimer", "", tef)
	
	If timer Is timerRecord Then
		timerRecord.cmsec = -250
		
		Select Case state
		Case stateRecMotion
			Call RecordWaypt(skapi.maplocCur, True)
		Case stateRecDoor
			Call RecordTraversingDoor(skapi.maplocCur)
		End Select
		
		itickRecord = itickRecord + 1
		Select Case itickRecord
		Case 2
			Call skapi.SetControlProperty(szPanelName, "btnRecord", "TextColor", clrBtnDisabled)
		Case 4
			Call skapi.SetControlProperty(szPanelName, "btnRecord", "TextColor", clrBtnFlash)
			itickRecord = 0
		End Select
	End If
	
	Call TraceExit("XrouteEdit.OnTimer", "")
End Function

Public Function OnHotkey(cmid, vk)
	Call TraceEnterEx("XrouteEdit.OnHotkey", SzQuote(cmid) + ", " + CStr(vk), tef)
	
	Select Case cmid
		
	Case cmidUseSelected
		Select Case state
		Case stateRecMotion
		'	If user is Using a door or portal, add it to the route.
			Dim aco : Set aco = skapi.acoSelected
			If Not (aco Is Nothing) Then
				If FAcoIsDoor(aco) Then
					Call RecordDoor()
				ElseIf FAcoIsPortal(aco) Then
					Call RecordPortal()
				End If
			End If
		End Select
		
	End Select
	
	Call TraceExit("XrouteEdit.OnHotkey", "")
End Function

Public Function OnEndPortalSelf()
	Call TraceEnterEx("XrouteEdit.OnEndPortalSelf", "", tef)
	
	Select Case state
	Case stateRecPortal
		Call RecordPortalExit(skapi.maplocCur)
	End Select
	
	Call TraceExit("XrouteEdit.OnEndPortalSelf", "")
End Function


'----------
' Privates
'----------

Private Property Get clrBtnEnabled()
	clrBtnEnabled = &H7BF3F7
End Property
Private Property Get clrBtnDisabled()
	clrBtnDisabled = &H808080
End Property
Private Property Get clrBtnFlash()
	clrBtnFlash = clrBtnEnabled
End Property

Private Property Get clrSel()
	clrSel = &HFFFFFF
End Property
Private Property Get clrUnsel()
	clrUnsel = &HC0C0C0
End Property

Private Property Get clrClean()
	clrClean = &H000000
End Property
Private Property Get clrDirty()
	clrDirty = &H000080
End Property

Private Property Get icolType()
	icolType = 0
End Property
Private Property Get icolIn()
	icolIn = 1
End Property
Private Property Get icolOut()
	icolOut = 2
End Property
Private Property Get icolWalk()
	icolWalk = 3
End Property

Private Property Get iconDoor()
	iconDoor = &H6000000 + &H1ED3
End Property
Private Property Get iconJump()
	iconJump = &H6000000 + &H1DA6
End Property
Private Property Get iconPortal()
	iconPortal = &H6000000 + &H139C
End Property
Private Property Get iconRecall()
	iconRecall = &H6000000 + &H139B
End Property
Private Property Get iconInclude()
	iconInclude = &H6000000 + &H11F7
End Property
Private Property Get iconCustom()
	iconCustom = &H6000000 + &H1F99
End Property


Private szPanelName
Private dictSzFEnabled

Private xdocRoutes
Private dictSzXroute

Private Property Get stateNormal()
	stateNormal = 0
End Property
Private Property Get statePlayback()
	statePlayback = 1
End Property
Private Property Get stateRecMotion()
	stateRecMotion = 2
End Property
Private Property Get stateRecDoor()
	stateRecDoor = 3
End Property
Private Property Get stateRecPortal()
	stateRecPortal = 4
End Property

Private state
Private timerRecord, itickRecord

Private iwayptSetWalk


Private Function Reset()
	Call TraceEnter("XrouteEdit.Reset", "")
	
	szPanelName = ""
	Set dictSzFEnabled = Nothing
	
	szFileRoutesMbr = ""
	fDirtyMbr = False
	Set xdocRoutes = Nothing
	Set dictSzXroute = Nothing
	
	Set xrouteCurMbr = Nothing
	iwayptSelMbr = iNil
	
	state = stateNormal
	Set timerRecord = Nothing
	
	Call TraceExit("XrouteEdit.Reset", "")
End Function

Private Function DispatchControlEvent(szControl, dictSzValue)
	Call TraceEnter("XrouteEdit.DispatchControlEvent", _
	  SzQuote(szControl) + ", " + SzQuote(dictSzValue(szControl)))
	Do
	
'	Ignore disabled controls.
	If Not dictSzFEnabled(szControl) Then Exit Do
	
	If state = statePlayback Then
	'	Most controls are disabled during playback.
		Select Case szControl
		Case "btnGo", "listWaypts", "lboxAttr"
		'	Do nothing.
		Case "btnStop"
			Call Xnav.StopNav()
		Case Else
			Call Xnav.StopNav()
			Call RaiseCmr("That command is not available during playback.")
		End Select
		Exit Do
	End If
	
	Dim szRoute
	Dim iwaypt, icol
	Dim szAttr, szValue
	
	Select Case szControl
		
	Case "btnNewFile"
		Call NewRouteFile()
		
	Case "btnOpen"
		Call OpenRouteFile()
		
	Case "btnSave"
		Call FQuerySave(True)
		
	Case "btnRevert"
		Call Revert()
		
	Case "btnQuit"
		Call Quit()
		
	Case "lboxRoute"
		szRoute = dictSzValue("lboxRoute")
		If szRoute <> "" Then
			Set xrouteCur = dictSzXroute(szRoute)
		Else
			Set xrouteCur = Nothing
		End If
		
	Case "btnAddRoute"
		Call NewRoute()
		
	Case "btnDelRoute"
		szRoute = dictSzValue("lboxRoute")
		If szRoute <> "" Then
			Call DeleteRoute(szRoute)
		End If
		
	Case "listWaypts"
		iwaypt = CLng(SzReplace(dictSzValue("listWaypts"), "(\d+),(\d+)", "$2"))
		icol = CLng(SzReplace(dictSzValue("listWaypts"), "(\d+),(\d+)", "$1"))
		Call TraceLine("iwaypt = " + CStr(iwaypt) + ", icol = " + CStr(icol))
		If icol = icolWalk Then
		'	Read checkbox state.
			Call skapi.AddHandler(evidOnControlProperty, Me)
			Call skapi.GetControlProperty(szPanelName, "listWaypts", _
			  "Data(" + CStr(icol) + ", " + CStr(iwaypt) + ")")
			iwayptSetWalk = iwaypt
		Else
			iwayptSel = iwaypt
		End If
		
	Case "lboxAttr"
		szAttr = dictSzValue("lboxAttr")
		Call ShowAttrValue(iwayptSel, szAttr)
		
	Case "editAttr"
		szAttr = dictSzValue("lboxAttr")
		szValue = dictSzValue("editAttr")
		Call SetAttrValue(iwayptSel, szAttr, szValue)
		
	Case "btnInsWaypt"
		Call InsertWaypt(dictSzValue("lboxTag"))
		
	Case "btnDelWaypt"
		Call DeleteWaypt()
		
	Case "btnRecord"
		Select Case state
		Case stateNormal
			Call RecordStart()
		Case stateRecMotion, stateRecDoor, stateRecPortal
			Call RecordStop()
		End Select
		
	Case "btnGo"
		Call PlayRoute(True)
		
	Case "btnStop"
		Select Case state
		Case statePlayback
			Call Xnav.StopNav()
		End Select
		
	Case "btnResume"
		Call PlayRoute(False)
		
	End Select
	
	Loop While False
	Call TraceExit("XrouteEdit.DispatchControlEvent", "")
End Function


Private Function NewRouteFile()
	Call TraceEnter("XrouteEdit.NewRouteFile", "")
	Do
	
	If Not FQuerySave(False) Then Exit Do
	
	szFileRoutes = ""
	fDirty = False
	Set xdocRoutes = Xnav.XdocNewXrouteFile()
	
	Call BuildRouteList("")
	
	Loop While False
	Call TraceExit("XrouteEdit.NewRouteFile", "")
End Function

Private Function OpenRouteFile()
	Call TraceEnter("XrouteEdit.NewRouteFile", "")
	Do
	
	If Not FQuerySave(False) Then Exit Do
	
	Dim szFilter
'	Is XrouteConvert loaded?
	If Eval("TypeName(XrouteConvert)") = "XrouteConvertCls" Then
	'	Yes; show both .xroute and .route files.
		szFilter = ".*\.x?route"
	Else
	'	No; show just .xroute files.
		szFilter = ".*\.xroute"
	End If
	
	Dim szFile : szFile = FileBrowser.SzGetFilename("", szFilter)
	If szFile = "" Then Exit Do
	
	Call LoadRouteFile(szFile)
	
	Loop While False
	Call TraceExit("XrouteEdit.NewRouteFile", "")
End Function

Private Function LoadRouteFile(szFile)
	Call TraceEnter("XrouteEdit.LoadRouteFile", SzQuote(szFile))
	Do
	
	If Not FQuerySave(False) Then Exit Do
	
	Call skapi.SetControlProperty(szPanelName, "textFile", "Text", "Loading...")
	Dim szFileT
	
	If FSuffixMatchSz(szFile, ".xroute") Then
		Call skapi.OutputLine("Reading " + szFile + "...")
		szFileT = szFile
		fDirty = False
		Set xdocRoutes = XdocLoadFile(szFile)
	Else
		Call skapi.OutputLine("Converting " + szFile + "...")
		szFileT = fso.BuildPath(fso.GetParentFolderName(szFile), _
		  fso.GetBaseName(szFile) + ".xroute")
		fDirty = True
		Set xdocRoutes = XrouteConvert.XdocReadFile(szFile)
	End If
	
	Call BuildRouteList("")
	
	szFileRoutes = szFileT
	
	Loop While False
	Call TraceExit("XrouteEdit.LoadRouteFile", "")
End Function

Private Function FQuerySave(fForceSave)
	Call TraceEnter("XrouteEdit.FQuerySave", CStr(fForceSave))
	
	Call EnsureNoRecord()
	
	If fDirty Or fForceSave Then
		Dim szFolder, szFile
		If szFileRoutes <> "" Then
			szFolder = fso.GetParentFolderName(szFileRoutes)
			szFile = fso.GetFileName(szFileRoutes)
		Else
			szFolder = fso.GetAbsolutePathName(".")
			szFile = ""
		End If
		szFile = SzGetInput("Save", "Press Enter to save route file:", szFile)
		If szFile <> "" Then
			If fso.GetExtensionName(szFile) = "" Then
				szFile = szFile + ".xroute"
			End If
			Call SaveFile(fso.BuildPath(szFolder, szFile))
			FQuerySave = True
		Else
			FQuerySave = False
		End If
	Else
		FQuerySave = True
	End If
	
	Call TraceExit("XrouteEdit.FQuerySave", CStr(FQuerySave))
End Function

Private Function SaveFile(szFile)
	Call TraceEnter("XrouteEdit.SaveFile", CStr(szFile))
	
	Set xdocRoutes = Xnav.XdocNewXrouteFile()
	
	If dictSzXroute.Count > 0 Then
		Dim rgxroute : rgxroute = dictSzXroute.Items
		Dim iroute, xroute
		For iroute = 0 To UBound(rgxroute)
			Set xroute = rgxroute(iroute)
			Call xdocRoutes.documentElement.AppendChild(xroute.ElemSave(xdocRoutes))
		Next
	End If
	
	Call AdjustElemIndent(xdocRoutes.documentElement, 0)
	Call xdocRoutes.Save(szFile)
	szFileRoutes = szFile
	fDirty = False
	
	Call skapi.OutputLine(fso.GetFileName(szFile) + " has been saved.")
	
	Call TraceExit("XrouteEdit.SaveFile", "")
End Function

Private Function Revert()
	Call TraceEnter("XrouteEdit.Revert", "")
	Do
	
	Call EnsureNoRecord()
	
	If Not FConfirm("DISCARD changes") Then Exit Do
	
	fDirty = False
	
	If szFileRoutes <> "" Then
		Call LoadRouteFile(szFileRoutes)
	Else
		Call NewRouteFile()
	End If
	
	Loop While False
	Call TraceExit("XrouteEdit.Revert", "")
End Function

Private Function Quit()
	Call TraceEnter("XrouteEdit.NewRouteFile", "")
	Do
	
	If Not FQuerySave(False) Then Exit Do
	
'	Remove the control handler.
	Call skapi.RemoveHandler(evidOnControlEvent, Me)
	
'	Take down the panel.
	Call skapi.RemoveControls(szPanelName)
	
	Call Reset()
	
	Loop While False
	Call TraceExit("XrouteEdit.NewRouteFile", "")
End Function


Private Function BuildRouteList(szRouteSel)
	Call TraceEnter("XrouteEdit.BuildRouteList", SzQuote(szRouteSel))
	
	Call EnableBtn("btnAddRoute", False)
	Call EnableBtn("btnDelRoute", False)
	
	Set xrouteCur = Nothing
	
'	Create a dictionary of routes in the given file.
	Set dictSzXroute = Xnav.DictSzXrouteFromXdoc(xdocRoutes)
	
'	Clear the route dropdown.
	Call skapi.GetControlProperty(szPanelName, "lboxRoute", "Clear()")
	Call EnableBtn("btnGo", False)
	
	Dim irouteSel : irouteSel = iNil
	
	If dictSzXroute.Count > 0 Then
		Dim rgxroute : rgxroute = dictSzXroute.Items
		Dim iroute : iroute = 0
		For iroute = 0 To UBound(rgxroute)
			Dim xroute : Set xroute = rgxroute(iroute)
			
		'	Add the route to the dropdown.
			Call skapi.GetControlProperty(szPanelName, "lboxRoute", _
			  "AddChoice(" + SzQuote(xroute.szName) + ")")
			
			If szRouteSel <> "" Then
				If FEqSzI(xroute.szName, szRouteSel) Then irouteSel = iroute
			Else
				If irouteSel = iNil Then irouteSel = iroute
			End If
		Next
	End If
	
	Call EnableBtn("btnAddRoute", True)
	
	If irouteSel <> iNil Then
		irouteCur = irouteSel
	End If
	
	Call TraceExit("XrouteEdit.BuildRouteList", "")
End Function


Private szFileRoutesMbr

Private Property Get szFileRoutes()
	szFileRoutes = szFileRoutesMbr
End Property

Private Property Let szFileRoutes(szNew)
	Call TraceEnter("XrouteEdit.szFileRoutes", " <- " + SzQuote(szFileRoutes))
	
	szFileRoutesMbr = szNew
	
	Dim szShow
	If szFileRoutesMbr <> "" Then
		szShow = fso.GetFileName(szFileRoutesMbr)
	Else
		szShow = "New route file"
	End If
	Call skapi.SetControlProperty(szPanelName, "textFile", "Text", szShow)
	
	Call TraceExit("XrouteEdit.szFileRoutes", "")
End Property


Private fDirtyMbr

Private Property Get fDirty()
	fDirty = fDirtyMbr
End Property

Private Property Let fDirty(fNew)
	Call TraceEnter("XrouteEdit.fDirty", " <- " + CStr(fNew))
	
	If fDirtyMbr <> fNew Then
		fDirtyMbr = fNew
		Dim clr
		If fDirtyMbr Then clr = clrDirty Else clr = clrClean
		Call skapi.SetControlProperty(szPanelName, "textFile", "TextColor", clr)
	End If
	
	Call TraceExit("XrouteEdit.fDirty", "")
End Property


Private Property Get irouteCur()
	Call TraceEnter("XrouteEdit.irouteCur", "")
	
	If xrouteCur Is Nothing Then
		irouteCur = iNil
	Else
		irouteCur = IrouteFromXroute(xrouteCur)
	End If
	
	Call TraceExit("XrouteEdit.irouteCur", CStr(irouteCur))
End Property

Private Property Let irouteCur(iroute)
	Call TraceEnter("XrouteEdit.irouteCur", " <- " + CStr(iroute))
	
	Call skapi.SetControlProperty(szPanelName, "lboxRoute", "Selected", CLng(iroute))
	Call EnableBtn("btnDelRoute", iroute <> iNil)
	
	Call TraceExit("XrouteEdit.irouteCur", "")
End Property


Private xrouteCurMbr

Private Property Get xrouteCur()
	Set xrouteCur = xrouteCurMbr
End Property

Private Property Set xrouteCur(xrouteNew)
	Call TraceEnter("XrouteEdit.xrouteCur", " <- " + SzTraceObj(xrouteNew))

	If Not (xrouteNew Is xrouteCurMbr) Then
		Call EnsureNoRecord()
		
		Set xrouteCurMbr = xrouteNew
		
		Call ShowRouteData(iNil)
		Call EnableBtn("btnDelRoute", Not (xrouteCurMbr Is Nothing))
		Call EnableBtn("btnGo", Not (xrouteCurMbr Is Nothing))
	End If
	
	Call TraceExit("XrouteEdit.xrouteCur", "")
End Property


Private iwayptSelMbr

Private Property Get iwayptSel()
	iwayptSel = iwayptSelMbr
End Property

Private Property Let iwayptSel(iwaypt)
	Call TraceEnter("XrouteEdit.iwayptSel", " <- " + CStr(iwaypt))
	
	If iwayptSelMbr <> iwaypt Then
		If iwayptSelMbr >= 0 Then
			Call skapi.SetControlProperty(szPanelName, "listWaypts", _
			  "Color(" + CStr(icolIn) + ", " + CStr(iwayptSelMbr) + ")", clrUnsel)
			Call skapi.SetControlProperty(szPanelName, "listWaypts", _
			  "Color(" + CStr(icolOut) + ", " + CStr(iwayptSelMbr) + ")", clrUnsel)
		End If
		
		iwayptSelMbr = iwaypt
		
		If iwayptSelMbr >= 0 Then
			Call skapi.SetControlProperty(szPanelName, "listWaypts", _
			  "Color(" + CStr(icolIn) + ", " + CStr(iwayptSelMbr) + ")", clrSel)
			Call skapi.SetControlProperty(szPanelName, "listWaypts", _
			  "Color(" + CStr(icolOut) + ", " + CStr(iwayptSelMbr) + ")", clrSel)
			Call EnableBtn("btnInsWaypt", True)
			Call EnableBtn("btnDelWaypt", iwayptSelMbr < xrouteCur.cwaypt)
			Call EnableBtn("btnRecord", True)
		Else
			Call EnableBtn("btnInsWaypt", False)
			Call EnableBtn("btnDelWaypt", False)
			Call EnableBtn("btnRecord", False)
		End If
		
		Call ShowWayptAttrs(iwayptSelMbr)
	End If
	
	Call TraceExit("XrouteEdit.iwayptSel", "")
End Property


Private Function NewRoute()
	Call TraceEnter("XrouteEdit.NewRoute", "")
	Do
	
	Call EnsureNoRecord()
	
	Dim szRoute : szRoute = SzGetInput("Input", "Type a route name and press Enter:", "")
	If szRoute = "" Then
		Call skapi.SetControlProperty(szPanelName, "lboxRoute", "Selected", CLng(-1))
		Exit Do
	End If
	
	If dictSzXroute.Exists(szRoute) Then
		Call RaiseCmr("There is already a route named " + szRoute + ".")
	End If
	
	Dim xroute : Set xroute = Xnav.XrouteNew()
	xroute.szName = szRoute
	
'	Add the route to the dictionary.
	Set dictSzXroute(szRoute) = xroute
	fDirty = True
	
'	Add the route to the dropdown.
	Call skapi.GetControlProperty(szPanelName, "lboxRoute", _
	  "AddChoice(" + SzQuote(szRoute) + ")")
	
	irouteCur = dictSzXroute.Count - 1
	
	Loop While False
	Call TraceExit("XrouteEdit.NewRoute", "")
End Function

Private Function DeleteRoute(szRoute)
	Call TraceEnter("XrouteEdit.DeleteRoute", SzQuote(szRoute))
	Do
	
	Call EnsureNoRecord()
	
	If Not FConfirm("DELETE " + szRoute) Then Exit Do
	
	Dim iroute : iroute = IrouteFromSz(szRoute)

'	Remove the route from the dictionary.
	Call dictSzXroute.Remove(szRoute)
	fDirty = True
	
	Set xrouteCur = Nothing
	irouteCur = iNil
	
'	Remove the route from the dropdown.
	Call skapi.GetControlProperty(szPanelName, "lboxRoute", _
	  "RemoveChoice(" + CStr(iroute) + ")")
	
	Loop While False
	Call TraceExit("XrouteEdit.DeleteRoute", "")
End Function


Private Function InsertWaypt(szTag)
	Call TraceEnter("XrouteEdit.InsertWaypt", SzQuote(szTag))
	
	Dim iwaypt : iwaypt = iwayptSel
	Dim wayptNew : Set wayptNew = Xnav.WayptFromTag(szTag)
	Call wayptNew.FromMaplocCur()
	Call xrouteCur.InsertWaypt(iwaypt, wayptNew)
	Call ShowWayptEx(iwaypt, True)
	fDirty = True
	
	Call TraceExit("XrouteEdit.InsertWaypt", "")
End Function

Private Function DeleteWaypt()
	Call TraceEnter("XrouteEdit.DeleteWaypt", "")
	
	Dim iwaypt : iwaypt = iwayptSel
	Call xrouteCur.DeleteWaypt(iwaypt)
	Call skapi.GetControlProperty(szPanelName, "listWaypts", "DeleteRow(" + CStr(iwaypt) + ")")
	iwayptSelMbr = iNil
	iwayptSel = iwaypt
	fDirty = True
	
	Call TraceExit("XrouteEdit.DeleteWaypt", "")
End Function

Private Function RecordStart()
	Call TraceEnter("XrouteEdit.RecordStart", "")
	
'	Call skapi.SetControlProperty(szPanelName, "btnRecord", "Text", "Done")
	Call skapi.SetControlProperty(szPanelName, "btnRecord", "TextColor", clrBtnFlash)
	
	Call skapi.OutputLine("Your movements are now being recorded.  " + _
	  "Click Record again to stop recording.", opmChatWnd)
	
	Set timerRecord = skapi.TimerNew()
	Call skapi.AddHandler(evidOnTimer, Me)
	timerRecord.cmsec = -250
	itickRecord = 0
	
	Call skapi.Hotkey(cmidUseSelected, True)
	Call skapi.AddHandler(evidOnHotkey, Me)
	
	state = stateRecMotion
	
	Call TraceExit("XrouteEdit.RecordStart", "")
End Function

Private Function RecordWaypt(maploc, fOptimize)
	Call TraceEnter("XrouteEdit.RecordWaypt", SzTraceMaploc(maploc) + ", " + CStr(fOptimize))
	Do
	
	Dim iwaypt : iwaypt = iwayptSel
	
	If fOptimize And iwaypt > 0 Then
	'	If we haven't moved very far from previous waypoint then skip it.
		Dim wayptPrev : Set wayptPrev = xrouteCur.rgwaypt(iwaypt - 1)
		Dim maplocPrev : Set maplocPrev = wayptPrev.maplocOut
		If maplocPrev.Dist2DToMaploc(maploc) < Xnav.distMeter Then Exit Do
		
	'	If there's at least two previous waypoints, see if we're still on that same line.
		If iwaypt > 1 And Xnav.FSimpleWaypt(wayptPrev) Then
			Dim wayptPrev2 : Set wayptPrev2 = xrouteCur.rgwaypt(iwaypt - 2)
			Dim maplocPrev2 : Set maplocPrev2 = wayptPrev2.maplocOut
			If Abs(Xnav.DheadNormalize(maplocPrev2.HeadToMaploc(maplocPrev) - _
			  maplocPrev.HeadToMaploc(maploc))) < 10 And _
			  maplocPrev2.Dist2DToMaploc(maploc) < 0.5 Then
			'	Traveling in a more or less straight line.
			'	Extend current leg by replacing previous waypoint.
				wayptPrev.fWalk = FWalking()
				Call wayptPrev.Init(maploc, -1)
				Call ShowWayptEx(iwaypt - 1, False)
				fDirty = True
				Exit Do
			End If
		End If
	End If
	
	Dim wayptNew : Set wayptNew = Xnav.WayptFromTag("Simple")
	wayptNew.fWalk = FWalking()
	Call wayptNew.Init(maploc, -1)
	Call xrouteCur.InsertWaypt(iwaypt, wayptNew)
	Call ShowWayptEx(iwaypt, True)
	fDirty = True
	
	Loop While False
	Call TraceExit("XrouteEdit.RecordWaypt", "")
End Function

Private Function RecordDoor()
	Call TraceEnter("XrouteEdit.RecordDoor", "")
	
	Dim iwaypt : iwaypt = iwayptSel
	Dim wayptDoor : Set wayptDoor = Xnav.WayptFromTag("Door")
	wayptDoor.fWalk = FWalking()
	Call wayptDoor.FromMaplocCur()
	Call xrouteCur.InsertWaypt(iwaypt, wayptDoor)
	Call ShowWayptEx(iwaypt, True)
	fDirty = True
	
	If state = stateRecMotion Then
		state = stateRecDoor
	End If
	
	Call TraceExit("XrouteEdit.RecordDoor", "")
End Function

Private Function RecordTraversingDoor(maploc)
	Call TraceEnter("XrouteEdit.RecordTraversingDoor", SzTraceMaploc(maploc))
	
'	Have we passed through the door yet?  I.e. is its maplocOut behind us?
	Dim iwaypt : iwaypt = iwayptSel - 1
	Dim wayptDoor : Set wayptDoor = xrouteCur.rgwaypt(iwaypt)
	Dim maplocOut : Set maplocOut = wayptDoor.maplocOut
	Dim headToDoor : headToDoor = maploc.HeadToMaploc(maplocOut)
	If Abs(Xnav.DheadNormalize(headToDoor - maplocOut.head)) > 90 Then
	'	Yes; resume normal recording.
		state = stateRecMotion
	End If
	
	Call TraceExit("XrouteEdit.RecordTraversingDoor", "")
End Function

Private Function RecordPortal()
	Call TraceEnter("XrouteEdit.RecordPortal", "")
	
	Dim iwaypt : iwaypt = iwayptSel
	Dim wayptPortal : Set wayptPortal = Xnav.WayptFromTag("Portal")
	wayptPortal.fWalk = FWalking()
	Call wayptPortal.FromMaplocCur()
	Call xrouteCur.InsertWaypt(iwaypt, wayptPortal)
	Call ShowWayptEx(iwaypt, True)
	fDirty = True
	
	If state = stateRecMotion Then
		Call skapi.AddHandler(evidOnEndPortalSelf, Me)
		state = stateRecPortal
	End If
	
	Call TraceExit("XrouteEdit.RecordPortal", "")
End Function

Private Function RecordPortalExit(maploc)
	Call TraceEnter("XrouteEdit.RecordPortalExit", SzTraceMaploc(maploc))
	
	Call skapi.RemoveHandler(evidOnEndPortalSelf, Me)
	
'	Record the precise portal exit coords.
	Dim iwaypt : iwaypt = iwayptSel - 1
	Dim wayptPortal : Set wayptPortal = xrouteCur.rgwaypt(iwaypt)
	Set wayptPortal.maplocOut = maploc
	Call ShowWayptEx(iwaypt, False)
	fDirty = True
	
'	Resume normal recording.
	state = stateRecMotion
	
	Call TraceExit("XrouteEdit.RecordPortalExit", "")
End Function

Private Function RecordStop()
	Call TraceEnter("XrouteEdit.RecordStop", "")
	
	Call skapi.RemoveHandler(evidOnEndPortalSelf, Me)
	
	Call skapi.RemoveHandler(evidOnTimer, Me)
	Set timerRecord = Nothing
	
	Call skapi.RemoveHandler(evidOnHotkey, Me)
	Call skapi.Hotkey(cmidUseSelected, False)
	
'	Call skapi.SetControlProperty(szPanelName, "btnRecord", "Text", "Record")
	Call skapi.SetControlProperty(szPanelName, "btnRecord", "TextColor", clrBtnEnabled)
	
	Call skapi.OutputLine("Your movements are no longer being recorded.", opmChatWnd)
	
	state = stateNormal
	
	Call TraceExit("XrouteEdit.RecordStop", "")
End Function

Private Function EnsureNoRecord()
	Call TraceEnter("XrouteEdit.EnsureNoRecord", "")
	
	If state >= stateRecMotion Then Call RecordStop()
	
	Call TraceExit("XrouteEdit.EnsureNoRecord", "")
End Function

Private Function FWalking()
	Call TraceEnter("XrouteEdit.FWalking", "")
	
	Dim fRunAsDefault : fRunAsDefault = (skapi.chop And chopRunAsDefault) <> 0
	Dim fHoldRun : fHoldRun = (skapi.KeyState(cmidHoldRun) = 2)
	FWalking = (fHoldRun = fRunAsDefault)
	
	Call TraceExit("XrouteEdit.FWalking", CStr(FWalking))
End Function


Private Function PlayRoute(fGo)
	Call TraceEnter("XrouteEdit.PlayRoute", CStr(fGo))
	
	Call EnsureNoRecord()
	
    Call EnableBtn("btnGo", False)
    Call EnableBtn("btnResume", False)
	
	Call EnableBtn("btnStop", True)
	
	Call Pauser.Clear()
	
	Call skapi.AddHandler(evidOnLocChangeSelf, Me)
	
	state = statePlayback

'	Error-handling boilerplate.
	Dim erh : Set erh = ErhBegin()
	On Error Resume Next
	
	If fGo Then
		iwayptSel = 0
		Call xrouteCur.Go()
	Else
		xrouteCur.iwaypt = iwayptSel
		Call xrouteCur.ResumeRoute()
	End If
	
'	Error-handling boilerplate.
	Call EndErh(erh)
	On Error Goto 0
	
	state = stateNormal
	
	Call skapi.RemoveHandler(evidOnLocChangeSelf, Me)
	
	Call EnableBtn("btnStop", False)
	
    Call EnableBtn("btnGo", True)
	
	iwayptSel = xrouteCur.iwaypt
	
	If erh.errno <> errnoOK Then
		Call erh.PassItOn()
    ElseIf xrouteCur.iwaypt = xrouteCur.cwaypt Then
		Call skapi.OutputLine("Done.")
	Else
		Call skapi.OutputLine("Route was interrupted.")
	    Call EnableBtn("btnResume", True)
	End If
	
	Call TraceExit("XrouteEdit.PlayRoute", "")
End Function


Private Function ShowRouteData(iwayptSelNew)
	Call TraceEnter("XrouteEdit.ShowRouteData", CStr(iwayptSelNew))
	
	iwayptSel = iNil
	
	Call skapi.GetControlProperty(szPanelName, "listWaypts", "Clear")
	
	If Not (xrouteCur Is Nothing) Then
		Dim iwaypt
		For iwaypt = 0 To xrouteCur.cwaypt - 1
			Call skapi.GetControlProperty(szPanelName, "listWaypts", "AddRow")
			Call ShowWaypt(iwaypt)
		Next
		
		Call skapi.GetControlProperty(szPanelName, "listWaypts", "AddRow")
		Call skapi.SetControlProperty(szPanelName, "listWaypts", _
		  "Data(" + CStr(icolType) + ", " + CStr(iwaypt) + ", 1)", 0)
		Call skapi.SetControlProperty(szPanelName, "listWaypts", _
		  "Color(" + CStr(icolIn) + ", " + CStr(iwaypt) + ")", clrUnsel)
		Call skapi.SetControlProperty(szPanelName, "listWaypts", _
		  "Data(" + CStr(icolIn) + ", " + CStr(iwaypt) + ")", "End")
		
		If iwayptSelNew >= 0 And iwayptSelNew <= xrouteCur.cwaypt Then
			iwayptSel = iwayptSelNew
			Call ScrollToWaypt(iwayptSel)
		Else
			iwayptSel = xrouteCur.cwaypt
		End If
	End If
	
	Call TraceExit("XrouteEdit.ShowRouteData", "")
End Function

Private Function ShowWayptEx(iwaypt, fInsert)
	Call TraceEnter("XrouteEdit.ShowWayptEx", CStr(iwaypt) + ", " + CStr(fInsert))
	
	If fInsert Then
		Call skapi.GetControlProperty(szPanelName, "listWaypts", "InsertRow(" + CStr(iwaypt) + ")")
		If iwayptSelMbr >= iwaypt Then iwayptSelMbr = iwayptSelMbr + 1
	End If
	
	Call ShowWaypt(iwaypt)
	Call ScrollToWaypt(iwaypt)
	
	Call TraceExit("XrouteEdit.ShowWayptEx", "")
End Function

Private Function ShowWaypt(iwaypt)
	Call TraceEnter("XrouteEdit.ShowWaypt", CStr(iwaypt))
	
	Dim waypt : Set waypt = xrouteCur.rgwaypt(iwaypt)
	
	Dim maplocIn : Set maplocIn = waypt.maplocIn
	Dim maplocOut : Set maplocOut = waypt.maplocOut
	Dim szIn, szOut
	If maplocIn Is Nothing Then szIn = "" Else szIn = maplocIn.sz(3)
	If maplocOut Is Nothing Then szOut = "" Else szOut = maplocOut.sz(3)
	
	Dim icon
	Select Case waypt.szTag
	Case "Simple"
		icon = 0
	Case "Door"
		icon = iconDoor
	Case "Jump"
		icon = iconJump
	Case "Portal"
		icon = iconPortal
	Case "Recall"
		icon = iconRecall
	Case "Include"
		icon = iconInclude
	Case Else
		icon = iconCustom
	End Select
	
	Call skapi.SetControlProperty(szPanelName, "listWaypts", _
	  "Data(" + CStr(icolType) + ", " + CStr(iwaypt) + ", 1)", icon)
	Call skapi.SetControlProperty(szPanelName, "listWaypts", _
	  "Color(" + CStr(icolIn) + ", " + CStr(iwaypt) + ")", clrUnsel)
	Call skapi.SetControlProperty(szPanelName, "listWaypts", _
	  "Data(" + CStr(icolIn) + ", " + CStr(iwaypt) + ")", szIn)
	Call skapi.SetControlProperty(szPanelName, "listWaypts", _
	  "Color(" + CStr(icolOut) + ", " + CStr(iwaypt) + ")", clrUnsel)
	Call skapi.SetControlProperty(szPanelName, "listWaypts", _
	  "Data(" + CStr(icolOut) + ", " + CStr(iwaypt) + ")", szOut)
	Call skapi.SetControlProperty(szPanelName, "listWaypts", _
	  "Data(" + CStr(icolWalk) + ", " + CStr(iwaypt) + ")", waypt.fWalk)
	
	If iwayptSelMbr = iwaypt Then iwayptSelMbr = iNil
	
	Call TraceExit("XrouteEdit.ShowWaypt", "")
End Function

Private Function ScrollToWaypt(iwaypt)
	Call TraceEnter("XrouteEdit.ScrollToWaypt", CStr(iwaypt))
	
	Dim iwayptTop
	If iwaypt < 2 Then
		iwayptTop = 0
	Else
		iwayptTop = iwaypt - 2
	End If
	Call skapi.GetControlProperty(szPanelName, "listWaypts", _
	  "JumpToPosition(" + CStr(iwayptTop) + ")")
	
	Call TraceExit("XrouteEdit.ScrollToWaypt", "")
End Function

Private Function ShowWayptAttrs(iwaypt)
	Call TraceEnter("XrouteEdit.ShowWayptAttrs", CStr(iwaypt))
	
	Call skapi.GetControlProperty(szPanelName, "lboxAttr", "Clear()")
	Call skapi.SetControlProperty(szPanelName, "editAttr", "Text", "")
	Call EnableEdit("editAttr", False)
	
	If iwaypt < 0 Then
		Call skapi.SetControlProperty(szPanelName, "textTag", "Text", "")
	ElseIf iwaypt < xrouteCur.cwaypt Then
		Dim elem
		Set elem = Xnav.ElemFromWaypt(xdocRoutes, xrouteCur.rgwaypt(iwaypt), True)
		
		Call skapi.SetControlProperty(szPanelName, "textTag", "Text", elem.tagName)
		
		Call AddElemAttrs(elem, "")
		
		Dim sel : Set sel = elem.SelectNodes("*")
		Dim elemChild
		For Each elemChild In sel
			Call AddElemAttrs(elemChild, elemChild.tagName + ".")
		Next
	Else
		Call skapi.SetControlProperty(szPanelName, "textTag", "Text", "")
	End If
	
	Call TraceExit("XrouteEdit.ShowWayptAttrs", "")
End Function

Private Function AddElemAttrs(elem, szPrefix)
	Call TraceEnter("XrouteEdit.AddElemAttrs", SzTraceElem(elem) + ", " + SzQuote(szPrefix))
	
	Dim coattr : Set coattr = elem.attributes
	Dim attr
	For Each attr In coattr
		If attr.Name <> "fWalk" Then
			Call skapi.GetControlProperty(szPanelName, "lboxAttr", _
			  "AddChoice(" + SzQuote(szPrefix + attr.Name) + ")")
		End If
	Next
	
	Call TraceExit("XrouteEdit.AddElemAttrs", "")
End Function

Private Function ShowAttrValue(iwaypt, szAttr)
	Call TraceEnter("XrouteEdit.ShowAttrValue", CStr(iwaypt) + ", " + SzQuote(szAttr))
	
	Dim szValue
	If iwaypt < 0 Then
		szValue = ""
	ElseIf iwaypt < xrouteCur.cwaypt And szAttr <> "" Then
		Dim waypt : Set waypt = xrouteCur.rgwaypt(iwaypt)
		Dim elem : Set elem = Xnav.ElemFromWaypt(xdocRoutes, waypt, True)
		
		Dim szTag, szAttrT, elemT
		Call SplitSzRev(szTag, szAttrT, szAttr, ".")
		If szTag = "" Then
			Set elemT = elem
		Else
			Set elemT = elem.SelectSingleNode(szTag)
		End If
		
		szValue = elemT.GetAttribute(szAttrT)
		Call EnableEdit("editAttr", True)
	Else
		szValue = ""
		Call EnableEdit("editAttr", False)
	End If
	Call skapi.SetControlProperty(szPanelName, "editAttr", "Text", szValue)
	
	Call TraceExit("XrouteEdit.ShowAttrValue", "")
End Function

Private Function SetAttrValue(iwaypt, szAttr, szValue)
	Call TraceEnter("XrouteEdit.SetAttrValue", _
	  CStr(iwaypt) + ", " + SzQuote(szAttr) + ", " + SzQuote(szValue))
	
	Dim waypt : Set waypt = xrouteCur.rgwaypt(iwaypt)
	Dim elem : Set elem = Xnav.ElemFromWaypt(xdocRoutes, waypt, True)
	
	Dim szTag, szAttrT, elemT
	Call SplitSzRev(szTag, szAttrT, szAttr, ".")
	If szTag = "" Then
		Set elemT = elem
	Else
		Set elemT = elem.SelectSingleNode(szTag)
	End If
	
	Call elemT.SetAttribute(szAttrT, szValue)
	
	Call waypt.FromXml(elem)
	Call ShowWaypt(iwaypt)
	iwayptSel = iwaypt
	fDirty = True
	
	Call TraceExit("XrouteEdit.SetAttrValue", "")
End Function

Private Function IrouteFromSz(szRoute)
	Call TraceEnter("XrouteEdit.IrouteFromSz", SzQuote(szRoute))
	
	IrouteFromSz = iNil
	
	Dim rgszRoute : rgszRoute = dictSzXroute.Keys
	Dim iroute
	For iroute = 0 To UBound(rgszRoute)
		If FEqSzI(rgszRoute(iroute), szRoute) Then
			IrouteFromSz = iroute
			Exit For
		End If
	Next
	
	Call TraceExit("XrouteEdit.IrouteFromSz", CStr(IrouteFromSz))
End Function

Private Function IrouteFromXroute(xroute)
	Call TraceEnter("XrouteEdit.IrouteFromXroute", SzTraceObj(xroute))
	
	IrouteFromXroute = iNil
	
	Dim rgxroute : rgxroute = dictSzXroute.Items
	Dim iroute
	For iroute = 0 To UBound(rgxroute)
		If rgxroute(iroute) Is xroute Then
			IrouteFromXroute = iroute
			Exit For
		End If
	Next
	
	Call TraceExit("XrouteEdit.IrouteFromXroute", CStr(IrouteFromXroute))
End Function


Private Function FConfirm(sz)
	Call TraceEnter("XrouteEdit.FConfirm", SzQuote(sz))
	
	FConfirm = SzGetInput("Confirm", "Type OK to " + sz + ".", "") = "OK"
	
	Call TraceExit("XrouteEdit.FConfirm", CStr(FConfirm))
End Function


Private Function EnableBtn(szBtn, fEnable)
	Call TraceEnter("XrouteEdit.EnableBtn", SzQuote(szBtn) + ", " + CStr(fEnable))
	
	Dim clr
	If fEnable Then clr = clrBtnEnabled Else clr = clrBtnDisabled
    Call skapi.SetControlProperty(szPanelName, szBtn, "TextColor", clr)
	dictSzFEnabled(szBtn) = fEnable
	
	Call TraceExit("XrouteEdit.EnableBtn", "")
End Function

Private Function EnableEdit(szEdit, fEnable)
	Call TraceEnter("XrouteEdit.EnableEdit", SzQuote(szEdit) + ", " + CStr(fEnable))
	
	dictSzFEnabled(szEdit) = fEnable
	
	Call TraceExit("XrouteEdit.EnableEdit", "")
End Function


End Class


' Instantiate the global object.
Public XrouteEdit
Set XrouteEdit = New XrouteEditCls


